import typing
from abjad.system.AbjadObject import AbjadObject
from .Part import Part
from .PartAssignment import PartAssignment
from .Section import Section


class PartManifest(AbjadObject):
    """
    Part manifest.

    ..  container:: example

        Initializes from parts:

        >>> part_manifest = abjad.PartManifest(
        ...    abjad.Part(section='BassClarinet', section_abbreviation='BCL'),
        ...    abjad.Part(section='Violin', section_abbreviation='VN'),
        ...    abjad.Part(section='Viola', section_abbreviation='VA'),
        ...    abjad.Part(section='Cello', section_abbreviation='VC'),
        ...    )
        >>> len(part_manifest)
        4

    ..  container:: example

        Initializes from orchestra sections:

        >>> part_manifest = abjad.PartManifest(
        ...    abjad.Section(
        ...         abbreviation='FL',
        ...         count=4,
        ...         name='Flute',
        ...         ),
        ...    abjad.Section(
        ...         abbreviation='OB',
        ...         count=3,
        ...         name='Oboe',
        ...         ),
        ...    abjad.Part(
        ...         section_abbreviation='EH',
        ...         section='EnglishHorn',
        ...         ),
        ...    abjad.Section(
        ...         abbreviation='VN-1',
        ...         count=18,
        ...         instrument='Violin',
        ...         name='FirstViolin',
        ...         ),
        ...    abjad.Section(
        ...         abbreviation='VN-2',
        ...         count=18,
        ...         instrument='Violin',
        ...         name='SecondViolin',
        ...         ),
        ...    )
        >>> len(part_manifest)
        44

    """

    ### CLASS VARIABLES ###

    __slots__ = (
        '_parts',
        '_sections',
        )

    ### INITIALIZER ###

    def __init__(self, *arguments):
        parts, sections = [], []
        for argument in arguments:
            if isinstance(argument, Part):
                parts.append(argument)
            elif isinstance(argument, Section):
                sections.append(argument)
                parts.extend(argument.parts)
            else:
                raise TypeError(f'must be part or section (not {argument}).')
        for i, part in enumerate(parts):
            number = i + 1
            part._number = number
        self._parts = parts
        self._sections = sections

    ### SPECIAL METHODS ###

    def __iter__(self) -> typing.Iterator:
        """
        Iterates parts in manifest.

        ..  container:: example

            >>> part_manifest = abjad.PartManifest(
            ...    abjad.Section(
            ...         abbreviation='FL',
            ...         count=4,
            ...         name='Flute',
            ...         ),
            ...    abjad.Section(
            ...         abbreviation='OB',
            ...         count=3,
            ...         name='Oboe',
            ...         ),
            ...    abjad.Part(
            ...         section_abbreviation='EH',
            ...         section='EnglishHorn',
            ...         ),
            ...    abjad.Section(
            ...         abbreviation='VN-1',
            ...         count=18,
            ...         instrument='Violin',
            ...         name='FirstViolin',
            ...         ),
            ...    abjad.Section(
            ...         abbreviation='VN-2',
            ...         count=18,
            ...         instrument='Violin',
            ...         name='SecondViolin',
            ...         ),
            ...    )

            >>> for part in part_manifest:
            ...     part
            ...
            Part(instrument='Flute', member=1, number=1, section='Flute', section_abbreviation='FL')
            Part(instrument='Flute', member=2, number=2, section='Flute', section_abbreviation='FL')
            Part(instrument='Flute', member=3, number=3, section='Flute', section_abbreviation='FL')
            Part(instrument='Flute', member=4, number=4, section='Flute', section_abbreviation='FL')
            Part(instrument='Oboe', member=1, number=5, section='Oboe', section_abbreviation='OB')
            Part(instrument='Oboe', member=2, number=6, section='Oboe', section_abbreviation='OB')
            Part(instrument='Oboe', member=3, number=7, section='Oboe', section_abbreviation='OB')
            Part(instrument='EnglishHorn', number=8, section='EnglishHorn', section_abbreviation='EH')
            Part(instrument='Violin', member=1, number=9, section='FirstViolin', section_abbreviation='VN-1', zfill=2)
            Part(instrument='Violin', member=2, number=10, section='FirstViolin', section_abbreviation='VN-1', zfill=2)
            Part(instrument='Violin', member=3, number=11, section='FirstViolin', section_abbreviation='VN-1', zfill=2)
            Part(instrument='Violin', member=4, number=12, section='FirstViolin', section_abbreviation='VN-1', zfill=2)
            Part(instrument='Violin', member=5, number=13, section='FirstViolin', section_abbreviation='VN-1', zfill=2)
            Part(instrument='Violin', member=6, number=14, section='FirstViolin', section_abbreviation='VN-1', zfill=2)
            Part(instrument='Violin', member=7, number=15, section='FirstViolin', section_abbreviation='VN-1', zfill=2)
            Part(instrument='Violin', member=8, number=16, section='FirstViolin', section_abbreviation='VN-1', zfill=2)
            Part(instrument='Violin', member=9, number=17, section='FirstViolin', section_abbreviation='VN-1', zfill=2)
            Part(instrument='Violin', member=10, number=18, section='FirstViolin', section_abbreviation='VN-1', zfill=2)
            Part(instrument='Violin', member=11, number=19, section='FirstViolin', section_abbreviation='VN-1', zfill=2)
            Part(instrument='Violin', member=12, number=20, section='FirstViolin', section_abbreviation='VN-1', zfill=2)
            Part(instrument='Violin', member=13, number=21, section='FirstViolin', section_abbreviation='VN-1', zfill=2)
            Part(instrument='Violin', member=14, number=22, section='FirstViolin', section_abbreviation='VN-1', zfill=2)
            Part(instrument='Violin', member=15, number=23, section='FirstViolin', section_abbreviation='VN-1', zfill=2)
            Part(instrument='Violin', member=16, number=24, section='FirstViolin', section_abbreviation='VN-1', zfill=2)
            Part(instrument='Violin', member=17, number=25, section='FirstViolin', section_abbreviation='VN-1', zfill=2)
            Part(instrument='Violin', member=18, number=26, section='FirstViolin', section_abbreviation='VN-1', zfill=2)
            Part(instrument='Violin', member=1, number=27, section='SecondViolin', section_abbreviation='VN-2', zfill=2)
            Part(instrument='Violin', member=2, number=28, section='SecondViolin', section_abbreviation='VN-2', zfill=2)
            Part(instrument='Violin', member=3, number=29, section='SecondViolin', section_abbreviation='VN-2', zfill=2)
            Part(instrument='Violin', member=4, number=30, section='SecondViolin', section_abbreviation='VN-2', zfill=2)
            Part(instrument='Violin', member=5, number=31, section='SecondViolin', section_abbreviation='VN-2', zfill=2)
            Part(instrument='Violin', member=6, number=32, section='SecondViolin', section_abbreviation='VN-2', zfill=2)
            Part(instrument='Violin', member=7, number=33, section='SecondViolin', section_abbreviation='VN-2', zfill=2)
            Part(instrument='Violin', member=8, number=34, section='SecondViolin', section_abbreviation='VN-2', zfill=2)
            Part(instrument='Violin', member=9, number=35, section='SecondViolin', section_abbreviation='VN-2', zfill=2)
            Part(instrument='Violin', member=10, number=36, section='SecondViolin', section_abbreviation='VN-2', zfill=2)
            Part(instrument='Violin', member=11, number=37, section='SecondViolin', section_abbreviation='VN-2', zfill=2)
            Part(instrument='Violin', member=12, number=38, section='SecondViolin', section_abbreviation='VN-2', zfill=2)
            Part(instrument='Violin', member=13, number=39, section='SecondViolin', section_abbreviation='VN-2', zfill=2)
            Part(instrument='Violin', member=14, number=40, section='SecondViolin', section_abbreviation='VN-2', zfill=2)
            Part(instrument='Violin', member=15, number=41, section='SecondViolin', section_abbreviation='VN-2', zfill=2)
            Part(instrument='Violin', member=16, number=42, section='SecondViolin', section_abbreviation='VN-2', zfill=2)
            Part(instrument='Violin', member=17, number=43, section='SecondViolin', section_abbreviation='VN-2', zfill=2)
            Part(instrument='Violin', member=18, number=44, section='SecondViolin', section_abbreviation='VN-2', zfill=2)

        """
        return iter(self.parts)

    def __len__(self) -> int:
        """
        Gets number of parts in manifest.

        ..  container:: example

            >>> part_manifest = abjad.PartManifest(
            ...    abjad.Part(section='BassClarinet', section_abbreviation='BCL'),
            ...    abjad.Part(section='Violin', section_abbreviation='VN'),
            ...    abjad.Part(section='Viola', section_abbreviation='VA'),
            ...    abjad.Part(section='Cello', section_abbreviation='VC'),
            ...    )
            >>> len(part_manifest)
            4

        ..  container:: example

            >>> part_manifest = abjad.PartManifest(
            ...    abjad.Section(
            ...         abbreviation='FL',
            ...         count=4,
            ...         name='Flute',
            ...         ),
            ...    abjad.Section(
            ...         abbreviation='OB',
            ...         count=3,
            ...         name='Oboe',
            ...         ),
            ...    abjad.Part(
            ...         section_abbreviation='EH',
            ...         section='EnglishHorn',
            ...         ),
            ...    abjad.Section(
            ...         abbreviation='VN-1',
            ...         count=18,
            ...         instrument='Violin',
            ...         name='FirstViolin',
            ...         ),
            ...    abjad.Section(
            ...         abbreviation='VN-2',
            ...         count=18,
            ...         instrument='Violin',
            ...         name='SecondViolin',
            ...         ),
            ...    )

            >>> len(part_manifest)
            44

        """
        return len(self.parts)

    ### PUBLIC PROPERTIES ###

    @property
    def parts(self) -> typing.List[Part]:
        """
        Gets parts in manifest.

        ..  container:: example

            >>> part_manifest = abjad.PartManifest(
            ...    abjad.Part(section='BassClarinet', section_abbreviation='BCL'),
            ...    abjad.Part(section='Violin', section_abbreviation='VN'),
            ...    abjad.Part(section='Viola', section_abbreviation='VA'),
            ...    abjad.Part(section='Cello', section_abbreviation='VC'),
            ...    )
            >>> for part in part_manifest:
            ...     part
            ...
            Part(instrument='BassClarinet', number=1, section='BassClarinet', section_abbreviation='BCL')
            Part(instrument='Violin', number=2, section='Violin', section_abbreviation='VN')
            Part(instrument='Viola', number=3, section='Viola', section_abbreviation='VA')
            Part(instrument='Cello', number=4, section='Cello', section_abbreviation='VC')

        ..  container:: example

            >>> part_manifest = abjad.PartManifest(
            ...    abjad.Section(
            ...         abbreviation='FL',
            ...         count=4,
            ...         name='Flute',
            ...         ),
            ...    abjad.Section(
            ...         abbreviation='OB',
            ...         count=3,
            ...         name='Oboe',
            ...         ),
            ...    abjad.Part(
            ...         section_abbreviation='EH',
            ...         section='EnglishHorn',
            ...         ),
            ...    abjad.Section(
            ...         abbreviation='VN-1',
            ...         count=18,
            ...         instrument='Violin',
            ...         name='FirstViolin',
            ...         ),
            ...    abjad.Section(
            ...         abbreviation='VN-2',
            ...         count=18,
            ...         instrument='Violin',
            ...         name='SecondViolin',
            ...         ),
            ...    )

            >>> for part in part_manifest.parts:
            ...     part
            ...
            Part(instrument='Flute', member=1, number=1, section='Flute', section_abbreviation='FL')
            Part(instrument='Flute', member=2, number=2, section='Flute', section_abbreviation='FL')
            Part(instrument='Flute', member=3, number=3, section='Flute', section_abbreviation='FL')
            Part(instrument='Flute', member=4, number=4, section='Flute', section_abbreviation='FL')
            Part(instrument='Oboe', member=1, number=5, section='Oboe', section_abbreviation='OB')
            Part(instrument='Oboe', member=2, number=6, section='Oboe', section_abbreviation='OB')
            Part(instrument='Oboe', member=3, number=7, section='Oboe', section_abbreviation='OB')
            Part(instrument='EnglishHorn', number=8, section='EnglishHorn', section_abbreviation='EH')
            Part(instrument='Violin', member=1, number=9, section='FirstViolin', section_abbreviation='VN-1', zfill=2)
            Part(instrument='Violin', member=2, number=10, section='FirstViolin', section_abbreviation='VN-1', zfill=2)
            Part(instrument='Violin', member=3, number=11, section='FirstViolin', section_abbreviation='VN-1', zfill=2)
            Part(instrument='Violin', member=4, number=12, section='FirstViolin', section_abbreviation='VN-1', zfill=2)
            Part(instrument='Violin', member=5, number=13, section='FirstViolin', section_abbreviation='VN-1', zfill=2)
            Part(instrument='Violin', member=6, number=14, section='FirstViolin', section_abbreviation='VN-1', zfill=2)
            Part(instrument='Violin', member=7, number=15, section='FirstViolin', section_abbreviation='VN-1', zfill=2)
            Part(instrument='Violin', member=8, number=16, section='FirstViolin', section_abbreviation='VN-1', zfill=2)
            Part(instrument='Violin', member=9, number=17, section='FirstViolin', section_abbreviation='VN-1', zfill=2)
            Part(instrument='Violin', member=10, number=18, section='FirstViolin', section_abbreviation='VN-1', zfill=2)
            Part(instrument='Violin', member=11, number=19, section='FirstViolin', section_abbreviation='VN-1', zfill=2)
            Part(instrument='Violin', member=12, number=20, section='FirstViolin', section_abbreviation='VN-1', zfill=2)
            Part(instrument='Violin', member=13, number=21, section='FirstViolin', section_abbreviation='VN-1', zfill=2)
            Part(instrument='Violin', member=14, number=22, section='FirstViolin', section_abbreviation='VN-1', zfill=2)
            Part(instrument='Violin', member=15, number=23, section='FirstViolin', section_abbreviation='VN-1', zfill=2)
            Part(instrument='Violin', member=16, number=24, section='FirstViolin', section_abbreviation='VN-1', zfill=2)
            Part(instrument='Violin', member=17, number=25, section='FirstViolin', section_abbreviation='VN-1', zfill=2)
            Part(instrument='Violin', member=18, number=26, section='FirstViolin', section_abbreviation='VN-1', zfill=2)
            Part(instrument='Violin', member=1, number=27, section='SecondViolin', section_abbreviation='VN-2', zfill=2)
            Part(instrument='Violin', member=2, number=28, section='SecondViolin', section_abbreviation='VN-2', zfill=2)
            Part(instrument='Violin', member=3, number=29, section='SecondViolin', section_abbreviation='VN-2', zfill=2)
            Part(instrument='Violin', member=4, number=30, section='SecondViolin', section_abbreviation='VN-2', zfill=2)
            Part(instrument='Violin', member=5, number=31, section='SecondViolin', section_abbreviation='VN-2', zfill=2)
            Part(instrument='Violin', member=6, number=32, section='SecondViolin', section_abbreviation='VN-2', zfill=2)
            Part(instrument='Violin', member=7, number=33, section='SecondViolin', section_abbreviation='VN-2', zfill=2)
            Part(instrument='Violin', member=8, number=34, section='SecondViolin', section_abbreviation='VN-2', zfill=2)
            Part(instrument='Violin', member=9, number=35, section='SecondViolin', section_abbreviation='VN-2', zfill=2)
            Part(instrument='Violin', member=10, number=36, section='SecondViolin', section_abbreviation='VN-2', zfill=2)
            Part(instrument='Violin', member=11, number=37, section='SecondViolin', section_abbreviation='VN-2', zfill=2)
            Part(instrument='Violin', member=12, number=38, section='SecondViolin', section_abbreviation='VN-2', zfill=2)
            Part(instrument='Violin', member=13, number=39, section='SecondViolin', section_abbreviation='VN-2', zfill=2)
            Part(instrument='Violin', member=14, number=40, section='SecondViolin', section_abbreviation='VN-2', zfill=2)
            Part(instrument='Violin', member=15, number=41, section='SecondViolin', section_abbreviation='VN-2', zfill=2)
            Part(instrument='Violin', member=16, number=42, section='SecondViolin', section_abbreviation='VN-2', zfill=2)
            Part(instrument='Violin', member=17, number=43, section='SecondViolin', section_abbreviation='VN-2', zfill=2)
            Part(instrument='Violin', member=18, number=44, section='SecondViolin', section_abbreviation='VN-2', zfill=2)

        ..  container:: example

            >>> abjad.Part(section='FirstViolin', member=18) in part_manifest.parts
            True

            >>> abjad.Part(section='FirstViolin', member=19) in part_manifest.parts
            False

        """
        return list(self._parts)

    @property
    def sections(self) -> typing.List[Section]:
        """
        Gets sections in manifest.

        ..  container:: example

            >>> part_manifest = abjad.PartManifest(
            ...    abjad.Part(section='BassClarinet', section_abbreviation='BCL'),
            ...    abjad.Part(section='Violin', section_abbreviation='VN'),
            ...    abjad.Part(section='Viola', section_abbreviation='VA'),
            ...    abjad.Part(section='Cello', section_abbreviation='VC'),
            ...    )
            >>> part_manifest.sections
            []

        ..  container:: example

            >>> part_manifest = abjad.PartManifest(
            ...    abjad.Section(
            ...         abbreviation='FL',
            ...         count=4,
            ...         name='Flute',
            ...         ),
            ...    abjad.Section(
            ...         abbreviation='OB',
            ...         count=3,
            ...         name='Oboe',
            ...         ),
            ...    abjad.Part(
            ...         section_abbreviation='EH',
            ...         section='EnglishHorn',
            ...         ),
            ...    abjad.Section(
            ...         abbreviation='VN-1',
            ...         count=18,
            ...         instrument='Violin',
            ...         name='FirstViolin',
            ...         ),
            ...    abjad.Section(
            ...         abbreviation='VN-2',
            ...         count=18,
            ...         instrument='Violin',
            ...         name='SecondViolin',
            ...         ),
            ...    )

            >>> for section in part_manifest.sections:
            ...     section
            ...
            Section(abbreviation='FL', count=4, instrument='Flute', name='Flute')
            Section(abbreviation='OB', count=3, instrument='Oboe', name='Oboe')
            Section(abbreviation='VN-1', count=18, instrument='Violin', name='FirstViolin')
            Section(abbreviation='VN-2', count=18, instrument='Violin', name='SecondViolin')

        ..  container:: example

            >>> section = abjad.Section(
            ...     abbreviation='VN-1',
            ...     count=18,
            ...     instrument='Violin',
            ...     name='FirstViolin',
            ...     )
            >>> section in part_manifest.sections
            True

            >>> section = abjad.Section(
            ...     abbreviation='VN-1',
            ...     count=36,
            ...     instrument='Violin',
            ...     name='FirstViolin',
            ...     )
            >>> section in part_manifest.sections
            False

        """
        return list(self._sections)

    ### PUBLIC METHODS ###

    def expand(self, part_assignment):
        """
        Expands ``part_assignment``.

        ..  container:: example

            >>> part_manifest = abjad.PartManifest(
            ...    abjad.Section(
            ...         abbreviation='FL',
            ...         count=4,
            ...         name='Flute',
            ...         ),
            ...    abjad.Section(
            ...         abbreviation='OB',
            ...         count=3,
            ...         name='Oboe',
            ...         ),
            ...    abjad.Part(
            ...         section_abbreviation='EH',
            ...         section='EnglishHorn',
            ...         ),
            ...    abjad.Section(
            ...         abbreviation='VN-1',
            ...         count=18,
            ...         instrument='Violin',
            ...         name='FirstViolin',
            ...         ),
            ...    abjad.Section(
            ...         abbreviation='VN-2',
            ...         count=18,
            ...         instrument='Violin',
            ...         name='SecondViolin',
            ...         ),
            ...    )

            >>> part_assignment = abjad.PartAssignment('Oboe')
            >>> for part in part_manifest.expand(part_assignment):
            ...     part
            ...
            Part(instrument='Oboe', member=1, number=5, section='Oboe', section_abbreviation='OB')
            Part(instrument='Oboe', member=2, number=6, section='Oboe', section_abbreviation='OB')
            Part(instrument='Oboe', member=3, number=7, section='Oboe', section_abbreviation='OB')

        """
        assert isinstance(part_assignment, PartAssignment)
        parts = []
        for part in self.parts:
            if part.section == part_assignment.section:
                if part_assignment.token is None:
                    parts.append(part)
                elif part.member in part_assignment.members:
                    parts.append(part)
        return parts



from ply import lex  # type: ignore
from abjad import exceptions
from abjad import scheme as abjad_scheme
from abjad.system import Parser


class SchemeParser(Parser):
    """
    SchemeParser` mimics how LilyPond's embedded Scheme parser behaves.

    It parses a single Scheme expression and then stops,
    by raising a ``SchemeParserFinishedError``.

    The parsed expression and its exact length in characters
    are cached on the ``SchemeParser`` instance.

    It is intended to be used only in conjunction with ``LilyPondParser``.
    """

    ### CLASS VARIABLES ###

    __slots__ = (
        'cursor',
        'cursor_end',
        'expression_depth',
        'result',
        'string_accumulator',
        )

    ### INITIALIZER ###

    def __init__(self, debug=False):
        self.cursor = None
        self.cursor_end = None
        self.expression_depth = None
        self.result = None
        self.string_accumulator = None
        Parser.__init__(self, debug=debug)

    ### PRIVATE METHODS ###

    def _setup(self):
        self.expression_depth = 0
        self.cursor = 0

    ### LEX SETUP ###

    A = r'[A-Za-z]'
    N = r'[0-9]'
    DIGIT = r'{}'.format(N)
    UNSIGNED = r'{}+'.format(N)
    HEX = r'(X|x)[A-Fa-f0-9]+'
    INT = r'(-?{})'.format(UNSIGNED)
    REAL = r'(({}\.{}*)|(-?\.{}+))'.format(INT, N, N)
    INITIAL = r'({}|!|\$|%|&|\*|/|<|>|\?|~|_|\^|:|=)'.format(A)
    SUBSEQUENT = r'({}|{}|\.|\+|-)'.format(INITIAL, N)
    #IDENTIFIER = r'({}{}*|\+|-|\.\.\.)'.format(INITIAL, SUBSEQUENT)

    # this has been rewritten to prevent Sphinx from complaining that it looks like bad ReST
    IDENTIFIER = r'([A-Za-z!\$%&\*/<>\?~_\^:=][A-Za-z0-9!\$%&\*/<>\?~_\^:=\.\+-]*|[\+-]|\.\.\.)'

    states = (
        ('quote', 'exclusive'),
    )

    tokens = (
        'AMPERSAND',
        'ASTERISK',
        'BOOLEAN',
        'CARAT',
        'COLON',
        'DECIMAL',
        'DOLLAR',
        'EQUALS',
        'EXCLAMATION',
        'HASH',
        'HEXADECIMAL',
        'IDENTIFIER',
        'INTEGER',
        'L_CARAT',
        'L_PAREN',
        'MINUS',
        'PERCENT',
        'PERIOD',
        'PLUS',
        'QUESTION',
        'QUOTE',
        'R_CARAT',
        'R_PAREN',
        'SLASH',
        'STRING',
        'TILDE',
        'UNDERSCORE',
    )

    ### LEX METHODS ###

    t_ignore = ''

    def t_whitespace(self, t):
        r'[ \t\r]+'
        self.cursor += len(t.value)
        t.cursor_end = self.cursor
        pass

    def t_BOOLEAN(self, t):
        r'\#(T|F|t|f)'
        self.cursor += len(t.value)
        t.cursor_end = self.cursor
        if t.value[-1].lower() == 't':
            t.value = True
        else:
            t.value = False
        return t

    @lex.TOKEN(HEX)
    def t_HEXADECIMAL(self, t):
        self.cursor += len(t.value)
        t.cursor_end = self.cursor
        t.value = int(t.value[2:], 16)
        return t

    @lex.TOKEN(REAL)
    def t_DECIMAL(self, t):
        self.cursor += len(t.value)
        t.cursor_end = self.cursor
        t.value = float(t.value)
        return t

    def t_HASH(self, t):
        r'\#'
        self.cursor += len(t.value)
        t.cursor_end = self.cursor
        return t

    @lex.TOKEN(INT)
    def t_INTEGER(self, t):
        self.cursor += len(t.value)
        t.cursor_end = self.cursor
        t.value = int(t.value)
        return t

    def t_L_PAREN(self, t):
        r'\('
        self.cursor += len(t.value)
        t.cursor_end = self.cursor
        self.expression_depth += 1
        return t

    def t_R_PAREN(self, t):
        r'\)'
        self.cursor += len(t.value)
        t.cursor_end = self.cursor
        return t

    def t_quote(self, t):
        r'\"'
        self.cursor += len(t.value)
        t.cursor_end = self.cursor
        t.lexer.push_state('quote')
        self.string_accumulator = ''
        pass

    def t_quote_440(self, t):
        r'''\\[nt\\'"]
        '''
        self.cursor += len(t.value)
        t.cursor_end = self.cursor
        self.string_accumulator += t.value
        pass

    def t_quote_443(self, t):
        r'[^\\""]+'
        self.cursor += len(t.value)
        t.cursor_end = self.cursor
        self.string_accumulator += t.value
        pass

    def t_quote_446(self, t):
        r'\"'
        self.cursor += len(t.value)
        t.cursor_end = self.cursor
        t.lexer.pop_state()
        t.type = 'STRING'
        t.value = self.string_accumulator
        self.string_accumulator = ''
        return t

    def t_quote_456(self, t):
        r'.'
        self.cursor += len(t.value)
        t.cursor_end = self.cursor
        self.string_accumulator += t.value
        pass

    @lex.TOKEN(IDENTIFIER)
    def t_IDENTIFIER(self, t):
        self.cursor += len(t.value)
        t.cursor_end = self.cursor
        return t

    def t_newline(self, t):
        r'\n+'
        self.cursor += len(t.value)
        t.cursor_end = self.cursor
        t.lexer.lineno += t.value.count('\n')

    def t_anything(self, t):
        r'.'
        self.cursor += len(t.value)
        t.cursor_end = self.cursor
        types = {
            '&': 'AMPERSAND',
            '*': 'ASTERISK',
            '^': 'CARAT',
            ':': 'COLON',
            '$': 'DOLLAR',
            '=': 'EQUALS',
            '!': 'EXCLAMATION',
            '<': 'L_CARAT',
            '-': 'MINUS',
            '%': 'PERCENT',
            '.': 'PERIOD',
            '+': 'PLUS',
            '?': 'QUESTION',
            "'": 'QUOTE',
            '>': 'R_CARAT',
            ')': 'R_PAREN',
            '/': 'SLASH',
            '~': 'TILDE',
            '_': 'UNDERSCORE',
        }
        if t.value in types:
            t.type = types[t.value]
            return t
        else:
            #t.lexer.skip(1)
            pass

    def t_error(self, t):
        self.cursor += len(t.value)
        t.cursor_end = self.cursor
        if self.debug:
            print(("SchemeParser-{}: Illegal character {!r}".format(
                id(self), t.value[0])))
        #t.lexer.skip(1)

    t_quote_error = t_error
    t_quote_ignore = t_ignore

    ### YACC SETUP ###

    start = 'program'

    ### YACC METHODS ###

    ### program ###

    """<program> : <form>*
    """

    def p_program__forms(self, p):
        """
        program : forms
        """
        p.slice[0].cursor_end = p.slice[-1].cursor_end
        p[0] = p[1]

    def p_forms__EMPTY(self, p):
        """
        forms :
        """
        p[0] = []

    def p_forms__forms__form(self, p):
        """
        forms : forms form
        """
        p.slice[0].cursor_end = p.slice[-1].cursor_end
        p[0] = p[1] + [p[2]]

    ### form ###

    """<form> : <definition> | <expression>
    """

    def p_form__expression(self, p):
        """
        form : expression
        """
        #print 'form : expression'
        #print p[1]
        p.slice[0].cursor_end = p.slice[-1].cursor_end
        p[0] = p[1]
        self.result = p[0]
        self.cursor_end = p.slice[0].cursor_end
        if self.debug:
            print('PARSED {!r}'.format(self.lexer.lexdata[:self.cursor_end]))
        raise exceptions.SchemeParserFinishedError

    ### definition ###

    '''<definition> : <variable definition>
    |   <syntax definition>
    |   (begin <definition>*)
    |   (let-syntax (<syntax binding>*) <definition>*)
    |   (letrec-syntax (<syntax binding>*) <definition>*)
    |   <derived definition>
    '''

    ### variable definition ###

    '''<variable definition> : (define <variable> <expression>)
    |   (define (<variable> <variable>*) <body>)
    |   (define (<variable> <variable>* . <variable>) <body>)
    '''

    ### variable ###

    '''<variable> : <identifier>
    '''

    def p_variable__IDENTIFIER(self, p):
        """
        variable : IDENTIFIER
        """
        #print 'variable : IDENTIFIER'
        #print p[1]
        p.slice[0].cursor_end = p.slice[-1].cursor_end
        p[0] = abjad_scheme.Scheme(p[1])

    ### body ###

    '''<body> : <definition>* <expression>+
    '''

    ### syntax definition ###

    '''<syntax definition> : (define-syntax <keyword> <transformer expression>)
    '''

    ### keyword ###

    '''<keyword> : <identifier>
    '''

    ### syntax binding ###

    '''<syntax binding> : (<keyword> <transformer expression>)
    '''

    ### expression ###

    '''<expression> : <constant>
    |   <variable>
    |   (quote <datum>) | ' <datum>
    |   (lambda <formals> <body>)
    |   (if <expression> <expression> <expression>) | (if <expression> <expression>)
    |   (set! <variable> <expression>)
    |   <application>
    |   (let-syntax (<syntax binding>*) <expression>+)
    |   (letrec-syntax (<syntax binding>*) <expression>+)
    |   <derived expression>
    '''

    def p_expression__variable(self, p):
        """
        expression : variable
        """
        #print 'expression : variable'
        #print p[1]
        p.slice[0].cursor_end = p.slice[-1].cursor_end
        p[0] = p[1]

    def p_expression__QUOTE__datum(self, p):
        """
        expression : QUOTE datum
        """
        #print 'expression : QUOTE datum'
        #print p[2]
        p.slice[0].cursor_end = p.slice[-1].cursor_end
        datum = p[2]
        if isinstance(datum, abjad_scheme.Scheme):
            if datum._quoting:
                datum._quoting = "'" + datum._quoting
            else:
                datum._quoting = "'"
            p[0] = datum
        else:
            p[0] = abjad_scheme.Scheme(datum, quoting="'")

    def p_expression__constant(self, p):
        """
        expression : constant
        """
        p.slice[0].cursor_end = p.slice[-1].cursor_end
        p[0] = p[1]

    ### constant ###

    '''<constant> : <boolean> | <number> | <character> | <string>
    '''

    def p_constant__boolean(self, p):
        """
        constant : boolean
        """
        p.slice[0].cursor_end = p.slice[-1].cursor_end
        p[0] = p[1]
        if self.expression_depth < 1:
            #print 'constant : boolean'
            #print p[1]
            self.result = p[0]
            self.cursor_end = p.slice[0].cursor_end
            raise exceptions.SchemeParserFinishedError

    def p_constant__number(self, p):
        """
        constant : number
        """
        p.slice[0].cursor_end = p.slice[-1].cursor_end
        p[0] = p[1]
        if self.expression_depth < 1:
            #print 'constant : number'
            #print p[1]
            self.result = p[0]
            self.cursor_end = p.slice[0].cursor_end
            raise exceptions.SchemeParserFinishedError

    def p_constant__string(self, p):
        """
        constant : string
        """
        p.slice[0].cursor_end = p.slice[-1].cursor_end
        p[0] = p[1]
        if self.expression_depth < 1:
            #print 'constant : string'
            #print p[1]
            self.result = p[0]
            self.cursor_end = p.slice[0].cursor_end
            raise exceptions.SchemeParserFinishedError

    ### formals ###

    '''<formals> : <variable> | (<variable>*) | (<variable>+ . <variable>)
    '''

    ### application ###

    '''<application> : (<expression> <expression>*)
    '''

    ### identifier ###

    '''<identifier> : <initial> <subsequent>* | + | - | ...
    '''

    ### initial ###

    '''<initial> : <letter> | ! | $ | % | & | * | / | : | < | = | > | ? | ~ | _ | ^
    '''

    ### subsequent ###

    '''<subsequent> : <initial> | <digit> | . | + | -
    '''

    ### letter ###

    '''<letter> : a | b | ... | z
    '''

    ### digit ###

    '''<digit> : 0 | 1 | ... | 9
    '''

    ### datum ###

    '''<datum> : <boolean> | <number> | <character> | <string> | <symbol> | <list> | <vector>
    '''

    def p_datum__constant(self, p):
        """
        datum : constant
        """
        p.slice[0].cursor_end = p.slice[-1].cursor_end
        p[0] = p[1]

    def p_datum__symbol(self, p):
        """
        datum : symbol
        """
        p.slice[0].cursor_end = p.slice[-1].cursor_end
        p[0] = p[1]

    #def p_datum__boolean(self, p):
    #    """datum : boolean"""
    #    p[0] = p[1]

    #def p_datum__number(self, p):
    #    """datum : number"""
    #    p[0] = p[1]

    #def p_datum__string(self, p):
    #    """datum : string"""
    #    p[0] = p[1]

    def p_datum__list(self, p):
        """
        datum : list
        """
        p.slice[0].cursor_end = p.slice[-1].cursor_end
        p[0] = p[1]

    def p_datum__vector(self, p):
        """
        datum : vector
        """
        p.slice[0].cursor_end = p.slice[-1].cursor_end
        p[0] = p[1]

    def p_data__EMPTY(self, p):
        """
        data :
        """
        p[0] = []

    def p_data__data__datum(self, p):
        """
        data : data datum
        """
        p.slice[0].cursor_end = p.slice[-1].cursor_end
        p[0] = p[1] + [p[2]]

    ### boolean ###

    '''<boolean> : #t | #f
    '''

    def p_boolean__BOOLEAN(self, p):
        """
        boolean : BOOLEAN
        """
        p.slice[0].cursor_end = p.slice[-1].cursor_end
        p[0] = p[1]

    ### number ###

    '''<number> : <num 2> | <num 8> | <num 10> | <num 16>
    '''

    def p_number__DECIMAL(self, p):
        """
        number : DECIMAL
        """
        p.slice[0].cursor_end = p.slice[-1].cursor_end
        p[0] = p[1]

    def p_number__HEXADECIMAL(self, p):
        """
        number : HEXADECIMAL
        """
        p.slice[0].cursor_end = p.slice[-1].cursor_end
        p[0] = p[1]

    def p_number__INTEGER(self, p):
        """
        number : INTEGER
        """
        p.slice[0].cursor_end = p.slice[-1].cursor_end
        p[0] = p[1]

    ### character ###

    '''<character> : #\ <any character> | #\newline | #\space
    '''

    ### string ###

    '''<string> : " <string character>* "
    '''

    def p_string__STRING(self, p):
        """
        string : STRING
        """
        p.slice[0].cursor_end = p.slice[-1].cursor_end
        p[0] = p[1]

    ### string character ###

    '''<string character> : \" | \\ | <any character other than " or \>
    '''

    ### symbol ###

    '''<symbol> : <identifier>
    '''

    def p_symbol__IDENTIFIER(self, p):
        """
        symbol : IDENTIFIER
        """
        #print 'symbol : IDENTIFIER'
        #print p[1]
        p.slice[0].cursor_end = p.slice[-1].cursor_end
        p[0] = p[1]

    ### vector ###

    '''<vector> : #(<datum>*)
    '''

    def p_vector__HASH__L_PAREN__data__R_PAREN(self, p):
        """
        vector : HASH L_PAREN data R_PAREN
        """
        p.slice[0].cursor_end = p.slice[-1].cursor_end
        p[0] = p[3]
        self.expression_depth -= 1

    ### list ###

    '''<list> : (<datum>*) | (<datum>+ . <datum>) | <abbreviation>
    '''

    def p_list__L_PAREN__data__R_PAREN(self, p):
        """
        list : L_PAREN data R_PAREN
        """
        p.slice[0].cursor_end = p.slice[-1].cursor_end
        p[0] = p[2]
        self.expression_depth -= 1

    def p_list__L_PAREN__data__datum__PERIOD__datum__R_PAREN(self, p):
        """
        list : L_PAREN data datum PERIOD datum R_PAREN
        """
        p.slice[0].cursor_end = p.slice[-1].cursor_end
        result = p[2] + [p[3]] + [p[5]]
        if len(result) == 2:
            p[0] = abjad_scheme.SchemePair(tuple(result))
        else:
            p[0] = abjad_scheme.Scheme(result)
        self.expression_depth -= 1

    ### abbreviation ###

    '''<abbreviation> : ' <datum> | ` <datum> | , <datum> | ,@ <datum>
    '''

    ### error ###

    def p_error(self, p):
        if p:
            if self.debug:
                print(("SchemeParser-{}: Syntax error at {!r}".format(
                    id(self), p.value)))
            self._parser.errok()
        else:
            if self.debug:
                print(("SchemeParser-{}: Syntax error at EOF".format(id(self))))

    ### PUBLIC PROPERTIES ###

    @property
    def lexer_rules_object(self):
        """
        Lexer rules object of Scheme parser.
        """
        return self

    @property
    def parser_rules_object(self):
        """
        Parser rules object of Scheme parser.
        """
        return self



import copy
import typing
from abjad import enums
from abjad.core.Leaf import Leaf
from abjad.core.Selection import Selection
from abjad.lilypondnames.LilyPondTweakManager import LilyPondTweakManager
from abjad.system.AbjadObject import AbjadObject
from abjad.system.FormatSpecification import FormatSpecification
from abjad.system.LilyPondFormatBundle import LilyPondFormatBundle
from abjad.system.LilyPondFormatManager import LilyPondFormatManager
from abjad.system.StorageFormatManager import StorageFormatManager
from abjad.system.Tag import Tag
from abjad.system.Tags import Tags
from abjad.system.Wrapper import Wrapper
from abjad.timespans import Timespan
from abjad.top.inspect import inspect
from abjad.top.override import override
from abjad.top.select import select
from abjad.top.setting import setting
from abjad.top.tweak import tweak
from abjad.utilities.Duration import Duration
abjad_tags = Tags()


class Spanner(AbjadObject):
    """
    Spanner.

    Any object that stretches horizontally and encompasses leaves.

    Usually at the context of the voice (not staff or higher).

    Examples include beams, slurs, hairpins and trills.
    """

    ### CLASS VARIABLES ###

    __slots__ = (
        '_contiguity_constraint',
        '_deactivate',
        '_ignore_attachment_test',
        '_ignore_before_attach',
        '_leaves',
        '_lilypond_setting_name_manager',
        '_tag',
        '_tweaks',
        '_wrappers',
        )

    _empty_chord = '<>'

    ### INITIALIZER ###

    def __init__(self) -> None:
        self._contiguity_constraint = 'logical voice'
        self._deactivate = None
        self._ignore_attachment_test = None
        self._ignore_before_attach = None
        self._leaves: typing.List[Leaf] = []
        self._lilypond_setting_name_manager = None
        self._tag = None
        self._tweaks = None
        self._wrappers: typing.List[Wrapper] = []

    ### SPECIAL METHODS ###

    def __contains__(self, argument) -> bool:
        """
        Is true when spanner contains ``argument``.
        """
        for leaf in self.leaves:
            if leaf is argument:
                return True
        else:
            return False

    #def __copy__(self, *arguments) -> 'Spanner':
    def __copy__(self, *arguments):
        """
        Copies spanner.

        Does not copy spanner leaves.
        """
        new = type(self)(*self.__getnewargs__())
        if getattr(self, '_lilypond_setting_name_manager', None) is not None:
            new._lilypond_setting_name_manager = copy.copy(setting(self))
        if getattr(self, '_tweaks', None) is not None:
            new._tweaks = copy.copy(tweak(self))
        self._copy_keywords(new)
        return new

    def __getitem__(self, argument) -> typing.Union[Leaf, Selection]:
        """
        Gets leaf or selection identified by ``argument``.
        """
        if isinstance(argument, slice):
            leaves = self.leaves.__getitem__(argument)
            return select(leaves)
        return self.leaves.__getitem__(argument)

    def __getnewargs__(self) -> typing.Tuple:
        """
        Gets new arguments.
        """
        return ()

    def __getstate__(self) -> dict:
        """
        Gets state of spanner.
        """
        state = {}
        for class_ in type(self).__mro__:
            for slot in getattr(class_, '__slots__', ()):
                state[slot] = getattr(self, slot, None)
        return state

    def __iter__(self) -> typing.Iterator:
        """
        Iterates leaves in spanner.
        """
        return iter(self.leaves)

    def __len__(self) -> int:
        """
        Gets number of leaves in spanner.
        """
        return len(self.leaves)

    def __lt__(self, argument) -> bool:
        """
        Is true when spanner is less than ``argument``.

        Trivial comparison to allow doctests to work.
        """
        assert isinstance(argument, Spanner), repr(argument)
        return repr(self) < repr(argument)

    ### PRIVATE METHODS ###

    def _add_direction(self, string):
        if getattr(self, 'direction', False):
            string = f'{self.direction} {string}'
        return string

    def _append(self, leaf):
        if self._ignore_attachment_test:
            pass
        elif not self._attachment_test(leaf):
            raise Exception(f'can not attach {self!r} to {leaf!r}.')
        if self._contiguity_constraint == 'logical voice':
            leaves = self[-1:] + [leaf]
            leaves = select(leaves)
            if not leaves.are_contiguous_logical_voice():
                raise Exception(type(self), leaves)
        leaf._append_spanner(self)
        self._leaves.append(leaf)

    def _at_least_two_leaves(self, argument):
        leaves = select(argument).leaves()
        if 1 < len(leaves):
            return True
        return [
            'Requires at least two leaves.',
            f'Not just {leaves[0]!r}.',
            ]

    def _attach(
        self,
        argument,
        deactivate=None,
        tag=None,
        ):
        assert not self, repr(self)
        assert isinstance(argument, Selection), repr(argument)
        assert argument.are_leaves(), repr(argument)
        self._extend(argument)
        self._deactivate = deactivate
        if tag is not None:
            tag = Tag(tag)
        self._tag = tag

    def _attachment_test(self, argument):
        return isinstance(argument, Leaf)

    def _attachment_test_all(self, argument):
        return True

    def _before_attach(self, argument):
        pass

    def _block_all_leaves(self):
        """
        Not composer-safe.
        """
        for leaf in self:
            self._block_leaf(leaf)

    def _block_leaf(self, leaf):
        """
        Not composer-safe.
        """
        leaf._remove_spanner(self)

    def _constrain_contiguity(self):
        """
        Not composer-safe.
        """
        self._contiguity_constraint = 'logical_voice'

    def _copy(self, leaves):
        """
        Returns copy of spanner with ``leaves``.

        ``leaves`` must already be contained in spanner.
        """
        my_leaf = self._leaves[:]
        self._leaves = []
        result = copy.copy(self)
        self._leaves = my_leaf
        for leaf in leaves:
            assert leaf in self
        for leaf in leaves:
            result._leaves.append(leaf)
        result._unblock_all_leaves()
        return result

    def _copy_keywords(self, new):
        pass

    def _detach(self):
        self._sever_all_leaves()

    def _extend(self, leaves):
        leaf_input = list(self[-1:])
        leaf_input.extend(leaves)
        leaf_input = select(leaf_input)
        if self._contiguity_constraint == 'logical voice':
            if not leaf_input.are_contiguous_logical_voice():
                message = f'{self!r} leaves must be contiguous:\n'
                message += f'  {leaf_input}'
                raise Exception(message)
        for leaf in leaves:
            self._append(leaf)

    def _fracture(self, i, direction=None):
        """
        Fractures spanner at ``direction`` of leaf at index ``i``.

        Valid values for ``direction`` are ``Left``, ``Right`` and ``None``.

        Set ``direction=None`` to fracture on both left and right sides.

        Returns tuple of original, left and right spanners.
        """
        if i < 0:
            i = len(self) + i
        if direction is enums.Left:
            return self._fracture_left(i)
        elif direction is enums.Right:
            return self._fracture_right(i)
        elif direction is None:
            left = self._copy(self[:i])
            right = self._copy(self[i + 1:])
            center = self._copy(self[i:i + 1])
            self._block_all_leaves()
            return self, left, center, right
        else:
            message = f'direction {direction!r} must be left, right or none.'
            raise ValueError(message)

    def _fracture_left(self, i):
        left = self._copy(self[:i])
        right = self._copy(self[i:])
        self._block_all_leaves()
        return self, left, right

    def _fracture_right(self, i):
        left = self._copy(self[:i + 1])
        right = self._copy(self[i + 1:])
        self._block_all_leaves()
        return self, left, right

    def _get_basic_lilypond_format_bundle(self, leaf):
        bundle = LilyPondFormatBundle()
        return bundle

    def _get_compact_summary(self):
        if not len(self):
            return ''
        elif 0 < len(self) <= 8:
            return ', '.join([str(_) for _ in self])
        else:
            left = ', '.join([str(_) for _ in self[:2]])
            right = ', '.join([str(_) for _ in self[-2:]])
            number = len(self) - 4
            middle = f', ... [{number}] ..., '
            return left + middle + right

    def _get_format_specification(self):
        agent = StorageFormatManager(self)
        names = list(agent.signature_keyword_names)
        if self._get_compact_summary() == '':
            values = []
        else:
            values = [self._get_compact_summary()]
        return FormatSpecification(
            client=self,
            repr_is_indented=False,
            repr_args_values=values,
            storage_format_kwargs_names=names,
            )

    def _get_lilypond_format_bundle(self, leaf):
        bundle = self._get_basic_lilypond_format_bundle(leaf)
        return bundle

    def _get_preprolated_duration(self):
        return sum([_._get_preprolated_duration() for _ in self])

    def _get_summary(self):
        if 0 < len(self):
            return ', '.join([str(_) for _ in self])
        else:
            return ' '

    def _get_timespan(self, in_seconds=False):
        if len(self):
            timespan_ = self[0]._get_timespan(in_seconds=in_seconds)
            start_offset = timespan_.start_offset
            timespan_ = self[-1]._get_timespan(in_seconds=in_seconds)
            stop_offset = timespan_.stop_offset
        else:
            start_offset = Duration(0)
            stop_offset = Duration(0)
        return Timespan(
            start_offset=start_offset,
            stop_offset=stop_offset,
            )

    def _index(self, leaf):
        return self._leaves.index(leaf)

    def _insert(self, i, leaf):
        """
        Not composer-safe.
        """
        if not isinstance(leaf, Leaf):
            raise Exception(f'spanners attach only to leaves: {leaf!s}.')
        leaf._append_spanner(self)
        self._leaves.insert(i, leaf)

    def _is_exterior_leaf(self, leaf):
        """
        Is true if leaf is first or last in spanner.
        Is true if next leaf or previous leaf is none.
        """
        if leaf is self[0]:
            return True
        elif leaf is self[-1]:
            return True
        elif not leaf._get_leaf(-1) and not leaf._get_leaf(1):
            return True
        else:
            return False

    def _is_interior_leaf(self, leaf):
        if leaf not in self.leaves:
            return False
        if len(self.leaves) < 3:
            return False
        leaf_count = len(self.leaves)
        first_index = 0
        last_index = leaf_count - 1
        leaf_index = list(self.leaves).index(leaf)
        if first_index < leaf_index < last_index:
            return True
        return False

    def _is_my_only(self, leaf):
        return len(self) == 1 and leaf is self[0]

    def _remove(self, leaf):
        """
        Not composer-safe.
        """
        self._sever_leaf(leaf)

    def _remove_leaf(self, leaf):
        """
        Not composer-safe.
        """
        for i, leaf_ in enumerate(self.leaves):
            if leaf_ is leaf:
                self._leaves.pop(i)
                break
        else:
            raise ValueError(f'{leaf!r} not in spanner.')

    def _sever_all_leaves(self):
        """
        Not composer-safe.
        """
        for i in reversed(range(len(self))):
            leaf = self[i]
            self._sever_leaf(leaf)

    def _sever_leaf(self, leaf):
        """
        Not composer-safe.
        """
        self._block_leaf(leaf)
        self._remove_leaf(leaf)

    def _start_offset_in_me(self, leaf):
        leaf_start_offset = inspect(leaf).timespan().start_offset
        self_start_offset = inspect(self).timespan().start_offset
        return leaf_start_offset - self_start_offset

    def _stop_command_string(self):
        return self._stop_command

    def _stop_offset_in_me(self, leaf):
        leaf_start_offset = self._start_offset_in_me(leaf)
        leaf_stop_offset = leaf_start_offset + leaf._get_duration()
        return leaf_stop_offset

    @staticmethod
    def _tag_hide(strings):
        return LilyPondFormatManager.tag(
            strings,
            deactivate=False,
            tag=abjad_tags.HIDE_TO_JOIN_BROKEN_SPANNERS,
            )

    @staticmethod
    def _tag_show(strings):
        return LilyPondFormatManager.tag(
            strings,
            deactivate=True,
            tag=abjad_tags.SHOW_TO_JOIN_BROKEN_SPANNERS,
            )

    def _tweaked_start_command_strings(self):
        strings = []
        contributions = tweak(self)._list_format_contributions()
        strings.extend(contributions)
        start_command = self._start_command
        start_command = self._add_direction(start_command)
        strings.append(start_command)
        return strings

    def _unblock_all_leaves(self):
        """
        Not composer-safe.
        """
        for leaf in self:
            self._unblock_leaf(leaf)

    def _unblock_leaf(self, leaf):
        """
        Not composer-safe.
        """
        leaf._append_spanner(self)

    def _unconstrain_contiguity(self):
        """
        Not composer-safe.
        """
        self._contiguity_constraint = None

    ### PUBLIC PROPERTIES ###

    @property
    def leaves(self) -> Selection:
        """
        Gets leaves in spanner.
        """
        for leaf in self._leaves:
            if not isinstance(leaf, Leaf):
                message = f'spanners attach only to leaves (not {leaf!s}).'
                raise Exception(message)
        return select(self._leaves)

    @property
    def tweaks(self) -> typing.Optional[LilyPondTweakManager]:
        """
        Gets tweaks.
        """
        return self._tweaks



import itertools
from abjad import exceptions
from abjad.mathtools.Ratio import Ratio
from abjad.utilities.Duration import Duration
from .Selection import Selection


class LogicalTie(Selection):
    """
    Logical tie of a component.

    ..  container:: example

        >>> staff = abjad.Staff("c' d' e' ~ e'")
        >>> abjad.show(staff) # doctest: +SKIP

        >>> abjad.inspect(staff[2]).logical_tie()
        LogicalTie([Note("e'4"), Note("e'4")])

    """

    ### CLASS VARIABLES ###

    __slots__ = ()

    ### SPECIAL METHODS ###

    def __getitem__(self, argument):
        """
        Gets ``argument``.

        Returns component or vanilla selection (not logical tie).
        """
        result = self.items.__getitem__(argument)
        if isinstance(result, tuple):
            result = Selection(result)
        return result

    ### PRIVATE METHODS ###

    def _add_or_remove_notes_to_achieve_written_duration(
        self, new_written_duration):
        import abjad
        new_written_duration = Duration(new_written_duration)
        maker = abjad.NoteMaker()
        if new_written_duration.is_assignable:
            self[0].written_duration = new_written_duration
            for leaf in self[1:]:
                parent = leaf._parent
                if parent:
                    index = parent.index(leaf)
                    del(parent[index])
            first = self[0]
            for spanner in first._get_spanners(abjad.Tie):
                spanner._sever_all_leaves()
        elif new_written_duration.has_power_of_two_denominator:
            durations = maker(0, [new_written_duration])
            for leaf, token in zip(self, durations):
                leaf.written_duration = token.written_duration
            if len(self) == len(durations):
                pass
            elif len(durations) < len(self):
                for leaf in self[len(durations):]:
                    parent = leaf._parent
                    if parent:
                        index = parent.index(leaf)
                        del(parent[index])
            elif len(self) < len(durations):
                for spanner in self[0]._get_spanners(abjad.Tie):
                    spanner._sever_all_leaves()
                difference = len(durations) - len(self)
                extra_leaves = self[0] * difference
                for extra_leaf in extra_leaves:
                    for spanner in extra_leaf._get_spanners():
                        spanner._remove(extra_leaf)
                extra_tokens = durations[len(self):]
                for leaf, token in zip(extra_leaves, extra_tokens):
                    leaf.written_duration = token.written_duration
                ties = self[-1]._get_spanners(abjad.Tie)
                if not ties:
                    tie = abjad.Tie()
                    if all(tie._attachment_test(_) for _ in self):
                        abjad.attach(tie, self.leaves)
                self[-1]._splice(extra_leaves, grow_spanners=True)
        else:
            durations = maker(0, new_written_duration)
            assert isinstance(durations[0], abjad.Tuplet)
            tuplet = durations[0]
            logical_tie = tuplet[0]._get_logical_tie()
            duration = logical_tie._get_preprolated_duration()
            self._add_or_remove_notes_to_achieve_written_duration(duration)
            multiplier = tuplet.multiplier
            tuplet = abjad.Tuplet(multiplier, [])
            abjad.mutate(self.leaves).wrap(tuplet)
        return self[0]._get_logical_tie()

    def _fuse_leaves_by_immediate_parent(self):
        result = []
        parts = self._get_leaves_grouped_by_immediate_parents()
        for part in parts:
            result.append(part._fuse())
        return result

    def _get_leaves_grouped_by_immediate_parents(self):
        import abjad
        result = []
        pairs_generator = itertools.groupby(self, lambda x: id(x._parent))
        for key, values_generator in pairs_generator:
            group = abjad.select(list(values_generator))
            result.append(group)
        return result

    def _scale(self, multiplier):
        new_duration = multiplier * self.written_duration
        return self._add_or_remove_notes_to_achieve_written_duration(
            new_duration)

    ### PUBLIC PROPERTIES ###

    @property
    def head(self):
        """
        Reference to element ``0`` in logical tie.

        Returns component.
        """
        if self.items:
            return self.items[0]

    @property
    def is_pitched(self):
        """
        Is true when logical tie head is a note or chord.

        Returns true or false.
        """
        import abjad
        return isinstance(self.head, (abjad.Note, abjad.Chord))

    @property
    def is_trivial(self):
        """
        Is true when length of logical tie is less than or equal to ``1``.

        Returns true or false.
        """
        return len(self) <= 1

    @property
    def leaves(self):
        """
        Gets leaves in logical tie.

        Returns selection.
        """
        import abjad
        try:
            tie = self[0]._get_spanner(prototype=abjad.Tie)
        except exceptions.MissingSpannerError:
            assert self.is_trivial
            return abjad.select(self[0])
        selection = tie.leaves
        assert isinstance(selection, abjad.Selection)
        return selection

    @property
    def tail(self):
        """
        Gets last leaf in logical tie.

        Returns leaf.
        """
        if self.items:
            return self.items[-1]

    @property
    def tie(self):
        """
        Gets tie spanner governing logical tie.

        Returns tie spanner.
        """
        import abjad
        return abjad.inspect(self[0]).spanner(abjad.Tie)

    @property
    def written_duration(self):
        """
        Sum of written duration of all components in logical tie.

        Returns duration.
        """
        return sum([_.written_duration for _ in self])

    ### PUBLIC METHODS ###

    def to_tuplet(self, proportions):
        r"""
        Changes logical tie to tuplet.

        ..  container:: example

            >>> staff = abjad.Staff(r"df'8 c'8 ~ c'16 cqs''4")
            >>> abjad.attach(abjad.Dynamic('p'), staff[0])
            >>> abjad.attach(abjad.DynamicTrend('<'), staff[0])
            >>> abjad.attach(abjad.Dynamic('f'), staff[-1])
            >>> abjad.override(staff).dynamic_line_spanner.staff_padding = 3
            >>> time_signature = abjad.TimeSignature((9, 16))
            >>> abjad.attach(time_signature, staff[0])
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                \with
                {
                    \override DynamicLineSpanner.staff-padding = #3
                }
                {
                    \time 9/16
                    df'8
                    \p
                    \<
                    c'8
                    ~
                    c'16
                    cqs''4
                    \f
                }

            >>> logical_tie = abjad.inspect(staff[1]).logical_tie()
            >>> logical_tie.to_tuplet([2, 1, 1, 1])
            Tuplet(Multiplier(3, 5), "c'8 c'16 c'16 c'16")

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                \with
                {
                    \override DynamicLineSpanner.staff-padding = #3
                }
                {
                    \time 9/16
                    df'8
                    \p
                    \<
                    \tweak text #tuplet-number::calc-fraction-text
                    \times 3/5 {
                        c'8
                        c'16
                        c'16
                        c'16
                    }
                    cqs''4
                    \f
                }

            >>> abjad.show(staff) # doctest: +SKIP

        ..  container:: example

            >>> staff = abjad.Staff(r"c'8 ~ c'16 cqs''4")
            >>> crescendo = abjad.Hairpin('p < f')
            >>> abjad.attach(crescendo, staff[:])
            >>> abjad.override(staff).dynamic_line_spanner.staff_padding = 3
            >>> time_signature = abjad.TimeSignature((7, 16))
            >>> abjad.attach(time_signature, staff[0])
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                \with
                {
                    \override DynamicLineSpanner.staff-padding = #3
                }
                {
                    \time 7/16
                    c'8
                    ~
                    \<
                    \p
                    c'16
                    cqs''4
                    \f
                }

        Returns tuplet.
        """
        import abjad
        proportions = Ratio(proportions)
        target_duration = self._get_preprolated_duration()
        prolated_duration = target_duration / sum(proportions.numbers)
        basic_written_duration = \
            prolated_duration.equal_or_greater_power_of_two
        written_durations = [
            _ * basic_written_duration for _ in proportions.numbers
            ]
        maker = abjad.NoteMaker()
        try:
            notes = [abjad.Note(0, _) for _ in written_durations]
        except exceptions.AssignabilityError:
            denominator = target_duration._denominator
            note_durations = [
                Duration(_, denominator)
                for _ in proportions.numbers
                ]
            notes = maker(0, note_durations)
        tuplet = abjad.Tuplet.from_duration(target_duration, notes)
        for leaf in self:
            for spanner in leaf._get_spanners(abjad.Tie):
                spanner._sever_all_leaves()
        abjad.mutate(self).replace(tuplet)
        return tuplet



"""
Tools for math.
"""
from .Infinity import Infinity
from .NegativeInfinity import NegativeInfinity
from .NonreducedFraction import NonreducedFraction
from .NonreducedRatio import NonreducedRatio
from .Ratio import Ratio
from .all_are_equal import all_are_equal
from .all_are_integer_equivalent import all_are_integer_equivalent
from .all_are_integer_equivalent_numbers import all_are_integer_equivalent_numbers
from .all_are_nonnegative_integer_equivalent_numbers import all_are_nonnegative_integer_equivalent_numbers
from .all_are_nonnegative_integer_powers_of_two import all_are_nonnegative_integer_powers_of_two
from .all_are_nonnegative_integers import all_are_nonnegative_integers
from .all_are_pairs_of_types import all_are_pairs_of_types
from .all_are_positive_integers import all_are_positive_integers
from .are_relatively_prime import are_relatively_prime
from .arithmetic_mean import arithmetic_mean
from .binomial_coefficient import binomial_coefficient
from .cumulative_products import cumulative_products
from .cumulative_sums import cumulative_sums
from .cumulative_sums_pairwise import cumulative_sums_pairwise
from .difference_series import difference_series
from .divisors import divisors
from .factors import factors
from .fraction_to_proper_fraction import fraction_to_proper_fraction
from .greatest_common_divisor import greatest_common_divisor
from .greatest_power_of_two_less_equal import greatest_power_of_two_less_equal
from .integer_equivalent_number_to_integer import integer_equivalent_number_to_integer
from .integer_to_base_k_tuple import integer_to_base_k_tuple
from .integer_to_binary_string import integer_to_binary_string
from .is_assignable_integer import is_assignable_integer
from .is_integer_equivalent import is_integer_equivalent
from .is_integer_equivalent_n_tuple import is_integer_equivalent_n_tuple
from .is_integer_equivalent_number import is_integer_equivalent_number
from .is_nonnegative_integer import is_nonnegative_integer
from .is_nonnegative_integer_equivalent_number import is_nonnegative_integer_equivalent_number
from .is_nonnegative_integer_power_of_two import is_nonnegative_integer_power_of_two
from .is_positive_integer import is_positive_integer
from .is_positive_integer_equivalent_number import is_positive_integer_equivalent_number
from .is_positive_integer_power_of_two import is_positive_integer_power_of_two
from .least_common_multiple import least_common_multiple
from .partition_integer_by_ratio import partition_integer_by_ratio
from .partition_integer_into_canonic_parts import partition_integer_into_canonic_parts
from .sign import sign
from .weight import weight
from .yield_all_compositions_of_integer import yield_all_compositions_of_integer



def all_are_equal(argument):
    """
    Is true when ``argument`` is an iterable collection of equal items.

    ..  container:: example

        >>> abjad.mathtools.all_are_equal([99, 99, 99, 99, 99, 99])
        True

        >>> abjad.mathtools.all_are_equal(17)
        False

    ..  container:: example

        Is true when ``argument`` is empty:

        >>> abjad.mathtools.all_are_equal([])
        True

    Returns true or false.
    """
    try:
        first_element = None
        for element in argument:
            if first_element is None:
                first_element = element
            else:
                if not element == first_element:
                    return False
        return True
    except TypeError:
        return False



from .Infinity import Infinity


class NegativeInfinity(Infinity):
    """
    Negative infinity.

    ..  container:: example

        All numbers compare greater than negative infinity:

        >>> NegativeInfinity < -9999999
        True

    ..  container:: example

        Negative infinity compares equal to itself:

        >>> NegativeInfinity == NegativeInfinity
        True

    ..  container:: example

        Negative infinity compares less than infinity:

        >>> NegativeInfinity < Infinity
        True

    Initializes as a system singleton at start-up.

    Available as a built-in after Abjad start.
    """

    ### CLASS VARIABLES ###

    __slots__ = ()

    ### INITIALIZER ###

    def __init__(self):
        self._value = float('-infinity')



import copy
from abjad import exceptions
from .Container import Container


class Measure(Container):
    r"""
    Measure.

    ..  container:: example

        >>> measure = abjad.Measure((4, 8), "c'8 d'8 e'8 f'8")
        >>> abjad.show(measure) # doctest: +SKIP

    ..  docs::

        >>> abjad.f(measure)
        {   % measure
            \time 4/8
            c'8
            d'8
            e'8
            f'8
        }   % measure

    """

    ### CLASS VARIABLES ###

    __slots__ = (
        '_always_format_time_signature',
        '_automatically_adjust_time_signature',
        '_measure_number',
        '_implicit_scaling',
        )

    ### INITIALIZER ###

    def __init__(
        self,
        time_signature=None,
        components=None,
        identifier='% measure',
        implicit_scaling=False,
        ):
        import abjad
        # set time signature adjustment before contents initialization
        self._automatically_adjust_time_signature = False
        time_signature = time_signature or abjad.TimeSignature((4, 4))
        time_signature = abjad.TimeSignature(time_signature)
        self.implicit_scaling = bool(implicit_scaling)
        Container.__init__(self, components, identifier=identifier)
        self._always_format_time_signature = False
        self._measure_number = None
        time_signature = abjad.TimeSignature(time_signature)
        abjad.attach(time_signature, self)

    ### SPECIAL METHODS ###

    def __copy__(self, *arguments):
        """
        Copies measure.

        Copies indicators.

        Does not copy spanners.

        Does not copy children.

        Returns new component.
        """
        import abjad
        time_signature = self.time_signature
        if time_signature is not None:
            abjad.detach(time_signature, self)
        new_measure = Container.__copy__(self)
        if time_signature is not None:
            abjad.attach(time_signature, self)
            abjad.detach(abjad.TimeSignature, new_measure)
            abjad.attach(copy.copy(time_signature), new_measure)
        return new_measure

    def __delitem__(self, i):
        r"""
        Deletes measure item ``i``.

        ..  container:: example

            >>> measure = abjad.Measure((4, 8), "c'8 d'8 e'8 f'8")
            >>> measure.automatically_adjust_time_signature = True
            >>> abjad.show(measure) # doctest: +SKIP

            >>> del(measure[1])
            >>> abjad.show(measure) # doctest: +SKIP

            ..  docs::

                {
                    \time 3/8
                    c'8
                    e'8
                    f'8
                }

        Returns none.
        """
        old_time_signature = self.time_signature
        old_denominator = getattr(old_time_signature, 'denominator', None)
        Container.__delitem__(self, i)
        self._conditionally_adjust_time_signature(old_denominator)

    def __getnewargs__(self):
        """
        Gets arguments for shallow copy.

        Returns tuple.
        """
        return (
            self.time_signature,
            [],
            self.identifier,
            self.implicit_scaling,
            )

    def __repr__(self):
        """
        Gets interpreter representation of measure.

        ..  container:: example

            >>> measure = abjad.Measure((3, 8), "c'8 d'8 e'8")
            >>> abjad.show(measure) # doctest: +SKIP

            >>> measure
            Measure((3, 8), "c'8 d'8 e'8")

        Returns string.
        """
        import abjad
        class_name = type(self).__name__
        indicator = self._get_indicator(abjad.TimeSignature)
        forced = indicator
        forced = forced.pair
        summary = self._get_contents_summary()
        if forced and len(self):
            if self.implicit_scaling:
                result = '{}({!s}, {!r}, implicit_scaling=True)'
                result = result.format(
                    class_name,
                    forced,
                    summary,
                    )
            else:
                result = f'{class_name}({forced!s}, {summary!r})'
            return result
        elif forced:
            if self.implicit_scaling:
                string = f'{class_name}({forced!s}, implicit_scaling=True)'
            else:
                string = f'{class_name}({forced!s})'
            return string
        else:
            if self.implicit_scaling:
                return f'{class_name}(implicit_scaling=True)'
            else:
                return f'{class_name}()'

    def __setitem__(self, i, argument):
        r"""
        Sets measure item ``i`` to ``argument``.

        ..  container:: example

            >>> measure = abjad.Measure((4, 8), "c'8 d'8 e'8 f'8")
            >>> measure.automatically_adjust_time_signature = True
            >>> abjad.show(measure) # doctest: +SKIP

            >>> measure[1] = abjad.Note("ds'8.")
            >>> abjad.show(measure) # doctest: +SKIP

            ..  docs::

                {
                    \time 9/16
                    c'8
                    ds'8.
                    e'8
                    f'8
                }

        Returns none.
        """
        old_time_signature = self.time_signature
        old_denominator = getattr(old_time_signature, 'denominator', None)
        Container.__setitem__(self, i, argument)
        self._conditionally_adjust_time_signature(old_denominator)

    ### PRIVATE METHODS ###

    def _all_contents_are_scalable_by_multiplier(self, multiplier):
        import abjad
        multiplier = abjad.Multiplier(multiplier)
        for component in self:
            if isinstance(component, abjad.Leaf):
                candidate_duration = multiplier * component.written_duration
                if not candidate_duration.is_assignable:
                    return False
        return True

    def _append_spacer_skip(self):
        import abjad
        if not self.is_underfull:
            return
        target_duration = self.time_signature.duration
        duration = self._get_duration()
        skip = abjad.Skip((1, 1))
        time_signature_multiplier = self.time_signature.implied_prolation
        new_duration = target_duration - duration
        new_multiplier = new_duration.__div__(time_signature_multiplier)
        abjad.attach(new_multiplier, skip)
        self.append(skip)

    def _as_graphviz_node(self):
        import abjad
        node = abjad.Component._as_graphviz_node(self)
        fraction = abjad.NonreducedFraction(
            self.time_signature.numerator,
            self.time_signature.denominator,
            )
        node[0].extend([
            uqbar.graphs.TableRow([
                uqbar.graphs.TableCell(
                    label=type(self).__name__,
                    attributes={'border': 0},
                    ),
                ]),
            uqbar.graphs.HRule(),
            uqbar.graphs.TableRow([
                uqbar.graphs.TableCell(
                    label=str(fraction),
                    attributes={'border': 0},
                    ),
                ]),
            ])
        return node

    def _check_duration(self):
        effective_time_signature = self.time_signature
        if (effective_time_signature.has_non_power_of_two_denominator and
            effective_time_signature.hide):
            message = 'can not hide time signature'
            message += ' with non-power-of-two denominator.'
            raise Exception(message)
        if effective_time_signature.duration < self._get_preprolated_duration():
            raise exceptions.OverfullContainerError
        if self._get_preprolated_duration() < effective_time_signature.duration:
            raise exceptions.UnderfullContainerError

    def _conditionally_adjust_time_signature(self, old_denominator):
        import abjad
        if self.automatically_adjust_time_signature:
            naive_time_signature = self._get_preprolated_duration()
            better_time_signature = \
                abjad.NonreducedFraction(naive_time_signature)
            better_time_signature = \
                better_time_signature.with_denominator(old_denominator)
            better_time_signature = abjad.TimeSignature(better_time_signature)
            abjad.detach(abjad.TimeSignature, self)
            abjad.attach(better_time_signature, self)

    @staticmethod
    def _duration_to_time_signature(
        duration,
        denominators=None,
        factor=None,
        ):
        import abjad
        duration = abjad.Duration(duration)
        if denominators is not None:
            if factor is not None:
                denominators = [
                    d for d in denominators
                    if factor in abjad.mathtools.factors(d)
                    ]
            for desired_denominator in sorted(denominators):
                nonreduced_fraction = abjad.NonreducedFraction(duration)
                candidate_pair = \
                    nonreduced_fraction.with_denominator(desired_denominator)
                if candidate_pair.denominator == desired_denominator:
                    return abjad.TimeSignature(candidate_pair)
        if factor is not None:
            if factor in abjad.mathtools.factors(duration.denominator):
                return abjad.TimeSignature(duration)
            else:
                time_signature_numerator = factor * duration.numerator
                time_signature_denominator = factor * duration.denominator
                return abjad.TimeSignature(
                    (time_signature_numerator, time_signature_denominator))
        else:
            return abjad.TimeSignature(duration)

    def _format_content_pieces(self):
        import abjad
        result = []
        if (self.has_non_power_of_two_denominator and
            type(self) is Measure and
            self.implicit_scaling):
            indent = abjad.LilyPondFormatManager.indent
            string = "{}\\scaleDurations #'({} . {}) {{"
            string = string.format(
                indent,
                self.implied_prolation.numerator,
                self.implied_prolation.denominator,
                )
            result.append(string)
            pieces = Container._format_content_pieces(self)
            pieces = [indent + _ for _ in pieces]
            result.extend(pieces)
            result.append(indent + '}')
        else:
            result.extend(
                Container._format_content_pieces(self))
        return result

    def _format_opening_slot(self, bundle):
        result = []
        result.append(('comments', bundle.opening.comments))
        result.append(('grob overrides', bundle.grob_overrides))
        result.append(('context settings', bundle.context_settings))
        result.append(('indicators', bundle.opening.indicators))
        return self._format_slot_contributions_with_indent(result)

    def _get_compact_representation(self):
        if not self:
            return f'| {self.time_signature!s} |'
        summary = self._get_contents_summary()
        return f'| {self.time_signature!s} {summary} |'

    def _get_format_specification(self):
        import abjad
        names = []
        if self.implicit_scaling:
            names.append('implicit_scaling')
        return abjad.FormatSpecification(
            client=self,
            repr_args_values=[
                self.time_signature.pair,
                self._get_contents_summary(),
                ],
            storage_format_args_values=[
                self.time_signature,
                self._get_contents_summary(),
                ],
            storage_format_kwargs_names=names,
            )

    def _get_lilypond_format(self):
        self._check_duration()
        return self._format_component()

    def _get_preprolated_duration(self):
        time_signature_prolation = 1
        if self.implicit_scaling:
            time_signature_prolation = self.time_signature.implied_prolation
        return time_signature_prolation * self._get_contents_duration()

    # TODO: see if self._scale can be combined with
    #       with self.scale_and_adjust_time_signature()
    def _scale(self, multiplier=None):
        import abjad
        if multiplier is None:
            return
        multiplier = abjad.Multiplier(multiplier)
        old_time_signature = self.time_signature
        if (abjad.mathtools.is_nonnegative_integer_power_of_two(multiplier) and
            1 <= multiplier):
            old_numerator = old_time_signature.numerator
            old_denominator = old_time_signature.denominator
            new_denominator = old_denominator // multiplier.numerator
            pair = (old_numerator, new_denominator)
            new_time_signature = abjad.TimeSignature(pair)
        else:
            old_denominator = old_time_signature.denominator
            old_duration = old_time_signature.duration
            new_duration = multiplier * old_duration
            new_time_signature = self._duration_to_time_signature(
                new_duration,
                [old_denominator],
                multiplier.denominator,
                )
        abjad.detach(abjad.TimeSignature, self)
        abjad.attach(new_time_signature, self)
        contents_multiplier_denominator = \
            abjad.mathtools.greatest_power_of_two_less_equal(
                multiplier.denominator)
        pair = (multiplier.numerator, contents_multiplier_denominator)
        contents_multiplier = abjad.Multiplier(*pair)
        self._scale_contents(contents_multiplier)

    def _scale_denominator(self, factor):
        import abjad
        # save old time signature duration
        old_time_signature_duration = self.time_signature.duration
        # find new time signature
        new_time_signature = self._duration_to_time_signature(
            old_time_signature_duration,
            factor=factor,
            )
        # scale contents of measures in argument
        multiplier = new_time_signature.implied_prolation.reciprocal
        self._scale(multiplier)
        # assign new time signature
        abjad.detach(abjad.TimeSignature, self)
        abjad.attach(new_time_signature, self)
        if new_time_signature.has_non_power_of_two_denominator:
            self.implicit_scaling = True

    ### PUBLIC PROPERTIES ###

    @property
    def always_format_time_signature(self):
        """
        Gets and sets flag to indicate whether time signature
        should appear in LilyPond format even when not expected.

        ..  container:: example

            >>> measure = abjad.Measure((4, 8), "c'8 d'8 e'8 f'8")
            >>> measure.always_format_time_signature
            False

        Set to true when necessary to print the same signature repeatedly.

        Defaults to false.

        Returns true or false.
        """
        return self._always_format_time_signature

    @always_format_time_signature.setter
    def always_format_time_signature(self, argument):
        assert isinstance(argument, bool)
        self._always_format_time_signature = argument

    @property
    def automatically_adjust_time_signature(self):
        """
        Gets and sets flag to indicate whether time signature
        should update automatically following mutation.

        ..  container:: example

            >>> measure = abjad.Measure((3, 4), "c' d' e'")
            >>> abjad.show(measure) # doctest: +SKIP

            >>> measure.automatically_adjust_time_signature = True
            >>> measure.append('r')
            >>> abjad.show(measure) # doctest: +SKIP

        Defaults to false.

        Returns true or false.
        """
        return self._automatically_adjust_time_signature

    @automatically_adjust_time_signature.setter
    def automatically_adjust_time_signature(self, argument):
        assert isinstance(argument, bool)
        self._automatically_adjust_time_signature = argument

    @property
    def has_non_power_of_two_denominator(self):
        """
        Is true when measure time signature denominator
        is not an integer power of 2.

        ..  container:: example

            >>> measure = abjad.Measure((5, 9), "c'8 d' e' f' g'")
            >>> measure.implicit_scaling = True
            >>> abjad.show(measure) # doctest: +SKIP

            >>> measure.has_non_power_of_two_denominator
            True

        Otherwise false:

        ..  container:: example

            >>> measure = abjad.Measure((5, 8), "c'8 d' e' f' g'")
            >>> abjad.show(measure) # doctest: +SKIP

            >>> measure.has_non_power_of_two_denominator
            False

        Returns true or false.
        """
        time_signature = self.time_signature
        return time_signature.has_non_power_of_two_denominator

    @property
    def has_power_of_two_denominator(self):
        """
        Is true when measure time signature denominator
        is an integer power of 2.

        ..  container:: example

            >>> measure = abjad.Measure((5, 8), "c'8 d' e' f' g'")
            >>> abjad.show(measure) # doctest: +SKIP

            >>> measure.has_power_of_two_denominator
            True

        Otherwise false:

        ..  container:: example

            >>> measure = abjad.Measure((5, 9), "c'8 d' e' f' g'")
            >>> measure.implicit_scaling = True
            >>> abjad.show(measure) # doctest: +SKIP

            >>> measure.has_power_of_two_denominator
            False

        Returns true or false.
        """
        return not self.has_non_power_of_two_denominator

    @property
    def implicit_scaling(self):
        """
        Is true when measure should scale contents.

        Returns true or false.
        """
        return self._implicit_scaling

    @implicit_scaling.setter
    def implicit_scaling(self, argument):
        assert isinstance(argument, bool)
        self._implicit_scaling = argument

    @property
    def implied_prolation(self):
        """
        Implied prolation of measure.

        ..  container:: example

            Measures with implicit scaling scale the duration of their
            contents:

            >>> measure = abjad.Measure((5, 12), "c'8 d'8 e'8 f'8 g'8")
            >>> measure.implicit_scaling = True
            >>> abjad.show(measure) # doctest: +SKIP

            >>> measure.implied_prolation
            Multiplier(2, 3)

            >>> for note in measure:
            ...     note, abjad.inspect(note).duration()
            (Note("c'8"), Duration(1, 12))
            (Note("d'8"), Duration(1, 12))
            (Note("e'8"), Duration(1, 12))
            (Note("f'8"), Duration(1, 12))
            (Note("g'8"), Duration(1, 12))

        ..  container:: example

            Measures without implicit scaling turned on do not
            scale the duration of their contents:

            >>> measure = abjad.Measure((5, 12), [])
            >>> measure.implicit_scaling = False

            >>> measure.implied_prolation
            Multiplier(1, 1)

        Returns positive multiplier.
        """
        import abjad
        if self.implicit_scaling:
            time_signature = self.time_signature
            return time_signature.implied_prolation
        return abjad.Multiplier(1)

    @property
    def is_full(self):
        """
        Is true when measure duration equals time signature duration.

        ..  container:: example

            >>> measure = abjad.Measure((4, 8), "c'8 d'8 e'8 f'8")
            >>> abjad.show(measure) # doctest: +SKIP

            >>> measure.is_full
            True

        Returns true or false.
        """
        return self._get_preprolated_duration() == self.target_duration

    @property
    def is_misfilled(self):
        """
        Is true when measure is either underfull or overfull.

        ..  container:: example

            >>> measure = abjad.Measure((3, 4), "c'4 d'4 e'4 f'4")
            >>> measure.is_misfilled
            True

        Otherwise false:

        ..  container:: example

            >>> measure = abjad.Measure((3, 4), "c' d' e'")
            >>> abjad.show(measure) # doctest: +SKIP

            >>> measure.is_misfilled
            False

        Returns true or false.
        """
        return not self.is_full

    @property
    def is_overfull(self):
        """I
        s true when measure duration is greater than time signature duration.

        ..  container:: example

            >>> measure = abjad.Measure((3, 4), "c'4 d' e' f'")

            >>> measure.is_overfull
            True

        Returns true or false.
        """
        return self.target_duration < self._get_preprolated_duration()

    @property
    def is_underfull(self):
        """
        Is true when measure duration is less than time signature duration.

        ..  container:: example

            >>> measure = abjad.Measure((3, 4), "c'4 d'")

            >>> measure.is_underfull
            True

        Returns true or false.
        """
        return self._get_preprolated_duration() < self.target_duration

    @property
    def measure_number(self):
        r"""
        Gets 1-indexed measure number.

        ..  container:: example

            >>> staff = abjad.Staff()
            >>> staff.append(abjad.Measure((3, 4), "c' d' e'"))
            >>> staff.append(abjad.Measure((2, 4), "f' g'"))
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff)
                \new Staff
                {
                    {   % measure
                        \time 3/4
                        c'4
                        d'4
                        e'4
                    }   % measure
                    {   % measure
                        \time 2/4
                        f'4
                        g'4
                    }   % measure
                }

            >>> staff[0].measure_number
            1

            >>> staff[1].measure_number
            2

        Returns positive integer.
        """
        self._update_now(offsets=True)
        return self._measure_number

    @property
    def target_duration(self):
        """
        Gets target duration of measure.

        Target duration of measure is always equal to duration
        of effective time signature.

        ..  container:: example

            >>> measure = abjad.Measure((3, 4), "c'4 d'4 e'4")
            >>> measure.target_duration
            Duration(3, 4)

        Returns duration.
        """
        return self.time_signature.duration

    @property
    def time_signature(self):
        """
        Gets effective time signature of measure.

        ..  container:: example

            >>> measure = abjad.Measure((3, 4), "c'4 d'4 e'4")
            >>> measure.time_signature
            TimeSignature((3, 4))

        Returns time signature or none.
        """
        import abjad
        return self._get_effective(abjad.TimeSignature)

    ### PUBLIC METHODS ###

    @classmethod
    def from_selections(class_, selections, time_signatures=None):
        """
        Makes a selection of measures from ``selections``.

        Returns selections.
        """
        import abjad
        assert len(selections)
        if not time_signatures:
            time_signatures = [_.duration() for _ in selections]
        assert len(selections) == len(time_signatures)
        durations = [abjad.inspect(_).duration() for _ in selections]
        assert durations == [abjad.Duration(_) for _ in time_signatures]
        maker = abjad.MeasureMaker()
        measures = maker(time_signatures)
        temporary_voice = abjad.Voice(measures)
        abjad.mutate(temporary_voice).replace_measure_contents(selections)
        temporary_voice[:] = []
        return measures

    # TODO: see if self._scale can be combined with
    #       with self.scale_and_adjust_time_signature()
    def scale_and_adjust_time_signature(self, multiplier=None):
        r"""
        Scales ``measure`` by ``multiplier`` and adjusts time signature.

        ..  container:: example

            Scales measure by non-power-of-two multiplier:

            >>> measure = abjad.Measure((3, 8), "c'8 d'8 e'8")
            >>> measure.implicit_scaling = True
            >>> abjad.show(measure) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(measure)
                {   % measure
                    \time 3/8
                    c'8
                    d'8
                    e'8
                }   % measure

            >>> measure.scale_and_adjust_time_signature(abjad.Multiplier(2, 3))
            >>> abjad.show(measure) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(measure)
                {   % measure
                    \time 3/12
                    \scaleDurations #'(2 . 3) {
                        c'8
                        d'8
                        e'8
                    }
                }   % measure

        Returns none.
        """
        import abjad
        if multiplier == 0:
            raise ZeroDivisionError
        old_time_signature = self.time_signature
        old_pair = old_time_signature.pair
        old_multiplier = old_time_signature.implied_prolation
        old_multiplier_pair = old_multiplier.pair
        multiplied_pair = abjad.NonreducedFraction(old_multiplier_pair)
        multiplied_pair = multiplied_pair.multiply_without_reducing(multiplier)
        multiplied_pair = multiplied_pair.pair
        reduced_pair = abjad.NonreducedFraction(old_multiplier_pair)
        reduced_pair = reduced_pair.multiply_with_cross_cancelation(multiplier)
        reduced_pair = reduced_pair.pair
        if reduced_pair != multiplied_pair:
            new_pair = abjad.NonreducedFraction(old_pair)
            new_pair = new_pair.multiply(multiplier, preserve_numerator=True)
            new_time_signature = abjad.TimeSignature(new_pair)
            abjad.detach(abjad.TimeSignature, self)
            abjad.attach(new_time_signature, self)
            remaining_multiplier = abjad.Multiplier(reduced_pair)
            if remaining_multiplier != abjad.Multiplier(1):
                self._scale_contents(remaining_multiplier)
        elif self._all_contents_are_scalable_by_multiplier(multiplier):
            self._scale_contents(multiplier)
            if (old_time_signature.has_non_power_of_two_denominator or
                not abjad.mathtools.is_nonnegative_integer_power_of_two(
                    multiplier)):
                new_pair = abjad.NonreducedFraction(old_pair)
                new_pair = new_pair.multiply_with_cross_cancelation(multiplier)
                new_pair = new_pair.pair
            # multiplier is a negative power of two, like 1/2, 1/4, etc.
            elif multiplier < abjad.Multiplier(0):
                new_pair = abjad.NonreducedFraction.multiply_without_reducing(
                    old_pair,
                    multiplier,
                    )
            # multiplier is a nonnegative power of two, like 0, 1, 2, 4, etc.
            elif abjad.Multiplier(0) < multiplier:
                new_pair = abjad.NonreducedFraction(old_pair)
                new_pair = new_pair.multiply(
                    multiplier,
                    preserve_numerator=True,
                    )
            elif multiplier == abjad.Multiplier(0):
                raise ZeroDivisionError
            new_time_signature = abjad.TimeSignature(new_pair)
            abjad.detach(abjad.TimeSignature, self)
            abjad.attach(new_time_signature, self)
            if new_time_signature.has_non_power_of_two_denominator:
                self.implicit_scaling = True
        else:
            new_pair = abjad.NonreducedFraction(old_pair)
            new_pair = new_pair.multiply(multiplier, preserve_numerator=True)
            new_time_signature = abjad.TimeSignature(new_pair)
            abjad.detach(abjad.TimeSignature, self)
            abjad.attach(new_time_signature, self)
            if new_time_signature.has_non_power_of_two_denominator:
                self.implicit_scaling = True
            implied_prolation = new_time_signature.implied_prolation
            remaining_multiplier = multiplier / implied_prolation
            if remaining_multiplier != abjad.Multiplier(1):
                self._scale_contents(remaining_multiplier)



def select(items=None):
    r"""
    Selects ``items`` or makes select expression.

    ..  container:: example

        Selects first two notes in staff:

        >>> staff = abjad.Staff("c'4 d' e' f'")
        >>> selection = abjad.select(staff[:2]).leaves(pitched=True)
        >>> for note in selection:
        ...     abjad.override(note).note_head.color = 'red'

        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(staff)
            \new Staff
            {
                \once \override NoteHead.color = #red
                c'4
                \once \override NoteHead.color = #red
                d'4
                e'4
                f'4
            }

    ..  container:: example

        Returns selection agent:

        >>> abjad.select(staff)
        Selection([Staff("c'4 d'4 e'4 f'4")])

        >>> abjad.select()
        abjad.select()

    """
    import abjad
    if items is None:
        return abjad.Expression().select()
    return abjad.Selection(items=items)



import collections
from abjad import markups
from abjad.system.AbjadValueObject import AbjadValueObject
from .Duplication import Duplication
from .Inversion import Inversion
from .Multiplication import Multiplication
from .Retrograde import Retrograde
from .Rotation import Rotation
from .Transposition import Transposition


class CompoundOperator(AbjadValueObject):
    """
    Compound operator.

    ..  container:: example

        Rotation followed by transposition:

        >>> operator = abjad.CompoundOperator()
        >>> operator = operator.rotate(n=1, stravinsky=True)
        >>> operator = operator.transpose(n=2)

        >>> str(operator)
        'T2rs1'

        >>> pitch_classes = abjad.PitchClassSegment([0, 1, 4, 7])
        >>> operator(pitch_classes)
        PitchClassSegment([2, 7, 8, 11])

    """

    ### CLASS VARIABLES ###

    __slots__ = (
        '_operators',
        '_show_identity_operators',
        )

    _publish_storage_format = True

    ### INITIALIZER ###

    def __init__(self, operators=None, *, show_identity_operators=None):
        if operators is not None:
            if not isinstance(operators, collections.Sequence):
                operators = (operators,)
            assert len(operators)
            operators = tuple(operators)
        self._operators = operators
        assert isinstance(show_identity_operators, (bool, type(None)))
        self._show_identity_operators = show_identity_operators

    ### SPECIAL METHODS ###

    def __add__(self, operator):
        """
        Composes compound operator and `operator`.

        ..  container:: example

            >>> operator = abjad.CompoundOperator()
            >>> operator = operator.transpose(n=1)
            >>> operator = operator.multiply(n=5)
            >>> str(operator)
            'M5T1'

        ..  container:: example

            >>> inversion = abjad.Inversion()
            >>> retrograde = abjad.Retrograde()
            >>> transposition = abjad.Transposition(n=1)

            >>> operator_1 = inversion + retrograde
            >>> str(operator_1)
            'IR'

            >>> operator_2 = inversion + transposition
            >>> str(operator_2)
            'IT1'

            >>> operator_3 = operator_1 + operator_2
            >>> str(operator_3)
            'IRIT1'

            >>> abjad.f(operator_3)
            abjad.CompoundOperator(
                operators=[
                    abjad.Transposition(
                        n=1,
                        ),
                    abjad.Inversion(),
                    abjad.Retrograde(),
                    abjad.Inversion(),
                    ],
                )

        Returns new compound operator.
        """
        operators = list(self.operators)
        if isinstance(operator, type(self)):
            operators[0:0] = operator.operators
        else:
            operators.insert(0, operator)
        result = type(self)()
        for operator in operators:
            result = result._with_operator(operator)
        return result

    def __call__(self, argument):
        """
        Calls compound operator on `argument`.

        ..  container:: example

            >>> operator = abjad.CompoundOperator()
            >>> operator = operator.rotate(n=1, stravinsky=True)
            >>> operator = operator.transpose(n=2)
            >>> str(operator)
            'T2rs1'

            >>> segment = abjad.PitchClassSegment([0, 1, 4, 7])
            >>> abjad.show(segment) # doctest: +SKIP

            >>> transform = operator(segment)
            >>> abjad.show(transform) # doctest: +SKIP

            >>> transform
            PitchClassSegment([2, 7, 8, 11])

        Returns new object with type equal to that of `argument`.
        """
        if self.operators is None:
            return argument
        for transform in self.operators:
            argument = transform(argument)
        return argument

    def __radd__(self, operator):
        """
        Composes `operator` and compound operator.

        ..  container

            >>> operator = abjad.CompoundOperator()
            >>> operator = operator.transpose(n=1)
            >>> operator = operator.multiply(n=5)
            >>> str(operator)
            'M5T1'

            >>> retrograde = abjad.Retrograde()
            >>> new_operator = retrograde + operator
            >>> str(new_operator)
            'RM5T1'

            >>> new_operator = operator + retrograde
            >>> str(new_operator)
            'M5T1R'

        Returns new compound operator.
        """
        operators = list(self.operators)
        if isinstance(operator, type(self)):
            operators.extend(operator.operators)
        else:
            operators.append(operator)
        result = type(self)()
        for operator in operators:
            result = result._with_operator(operator)
        return result

    def __str__(self):
        """
        Gets string representation of compound operator.

        ..  container:: example

            Gets string:

            >>> operator = abjad.CompoundOperator()
            >>> operator = operator.rotate(n=1)
            >>> operator = operator.transpose(n=2)

            >>> str(operator)
            'T2r1'

        ..  container:: example

            Gets string of empty operator:

            >>> operator = abjad.CompoundOperator()

            >>> str(operator)
            ''

        Returns string.
        """
        result = []
        operators = self.operators or []
        for operator in reversed(operators):
            if operator._is_identity_operator():
                if self.show_identity_operators:
                    result.append(str(operator))
            else:
                result.append(str(operator))
        result = ''.join(result)
        return result

    ### PRIVATE METHODS ###

    @staticmethod
    def _compose_operators(operator_1, operator_2):
        if isinstance(operator_1, CompoundOperator):
            result = operator_1.__add__(operator_2)
        elif isinstance(operator_2, CompoundOperator):
            result = operator_2.__radd__(operator_1)
        else:
            result = CompoundOperator()
            result = result._with_operator(operator_2)
            result = result._with_operator(operator_1)
        return result

    def _get_markup(self, direction=None):
        markups = []
        operators = self.operators or []
        for operator in operators:
            markup = operator._get_markup(direction=direction)
            if (self.show_identity_operators or
                not operator._is_identity_operator()):
                markups.append(markup)
        if len(markups) == 0:
            return
        elif len(markups) == 1:
            markup = markups[0]
        else:
            markup = markups.Markup.concat(markups, direction=direction)
        return markup

    def _with_operator(self, operator):
        operators = self.operators or []
        operators = operators + [operator]
        return type(self)(
            operators,
            show_identity_operators=self.show_identity_operators,
            )

    ### PUBLIC PROPERTIES ###

    @property
    def operators(self):
        """
        Gets operators.

        ..  container:: example

            Gets operators:

            >>> operator = abjad.CompoundOperator()
            >>> operator = operator.rotate(n=1)
            >>> operator = operator.transpose(n=2)

            >>> for operator_ in operator.operators:
            ...     operator_
            ...
            Rotation(n=1)
            Transposition(n=2)

        Returns list of operators.
        """
        if self._operators is not None:
            return list(self._operators)

    @property
    def show_identity_operators(self):
        """
        Is true when string representation of operator should show identity
        operators.

        ..  container:: example

            Does not show identity operators:

            >>> operator = abjad.CompoundOperator()
            >>> operator = operator.transpose(n=0)
            >>> operator = operator.multiply(n=5)

            >>> str(operator)
            'M5'

        ..  container:: example

            Shows identity operators:

            >>> operator = abjad.CompoundOperator(
            ...     show_identity_operators=True,
            ...     )
            >>> operator = operator.transpose(n=0)
            >>> operator = operator.multiply(n=5)

            >>> str(operator)
            'M5T0'

        Defaults to none.

        Set to true, false or none.

        Returns true, false or none.
        """
        return self._show_identity_operators

    ### PUBLIC METHODS ###

    def duplicate(self, counts=None, indices=None, period=None):
        """
        Configures compound operator to duplicate pitches by `counts`, with
        optional `indices` and `period`.

        ..  container:: example

            >>> operator = abjad.CompoundOperator()
            >>> operator = operator.duplicate(counts=1)
            >>> abjad.f(operator)
            abjad.CompoundOperator(
                operators=[
                    abjad.Duplication(
                        counts=1,
                        ),
                    ],
                )

        Returns new compound operator.
        """
        operator = Duplication(
            counts=counts,
            indices=indices,
            period=period,
            )
        return self._with_operator(operator)

    def invert(self, axis=None):
        """
        Configures compound operator to invert pitches about `axis`.

        ..  container:: example

            >>> operator = abjad.CompoundOperator()
            >>> operator = operator.invert(axis=2)
            >>> abjad.f(operator)
            abjad.CompoundOperator(
                operators=[
                    abjad.Inversion(
                        axis=abjad.NamedPitch("d'"),
                        ),
                    ],
                )

        Returns new compound operator.
        """
        operator = Inversion(axis=axis)
        return self._with_operator(operator)

    def multiply(self, n=1):
        """
        Configures compound operator to multiply pitch-classes by index `n`.

        ..  container:: example

            >>> operator = abjad.CompoundOperator()
            >>> operator = operator.multiply(n=3)
            >>> abjad.f(operator)
            abjad.CompoundOperator(
                operators=[
                    abjad.Multiplication(
                        n=3,
                        ),
                    ],
                )

        Returns new compound operator.
        """
        operator = Multiplication(n=n)
        return self._with_operator(operator)

    def retrograde(self, period=None):
        """
        Configures compound operator to retrograde pitches.

        ..  container:: example

            >>> operator = abjad.CompoundOperator()
            >>> operator = operator.retrograde()
            >>> abjad.f(operator)
            abjad.CompoundOperator(
                operators=[
                    abjad.Retrograde(),
                    ],
                )

        Returns new compound operator.
        """
        operator = Retrograde(period=period)
        return self._with_operator(operator)

    def rotate(self, n=0, period=None, stravinsky=None):
        """
        Configures compound operator to rotate pitches by index `n`.

        ..  container:: example

            >>> operator = abjad.CompoundOperator()
            >>> operator = operator.rotate(n=-1)
            >>> abjad.f(operator)
            abjad.CompoundOperator(
                operators=[
                    abjad.Rotation(
                        n=-1,
                        ),
                    ],
                )

        Returns new compound operator.
        """
        operator = Rotation(
            n=n,
            period=period,
            stravinsky=stravinsky,
            )
        return self._with_operator(operator)

    def transpose(self, n=0):
        """
        Configures compound operator to transpose pitches by index `n`.

        ..  container:: example

            >>> operator = abjad.CompoundOperator()
            >>> operator = operator.transpose(n=1)

            >>> abjad.f(operator)
            abjad.CompoundOperator(
                operators=[
                    abjad.Transposition(
                        n=1,
                        ),
                    ],
                )

        Returns new compound operator.
        """
        operator = Transposition(n=n)
        return self._with_operator(operator)



from .Context import Context


class Voice(Context):
    r"""
    Voice.

    ..  container:: example

        >>> voice = abjad.Voice("c'8 d'8 e'8 f'8")
        >>> abjad.show(voice) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(voice)
            \new Voice
            {
                c'8
                d'8
                e'8
                f'8
            }

    """

    ### CLASS VARIABLES ###

    __documentation_section__ = 'Contexts'

    __slots__ = ()

    _default_lilypond_type = 'Voice'

    ### INITIALIZER ###

    def __init__(
        self,
        components=None,
        lilypond_type='Voice',
        is_simultaneous=None,
        name=None,
        ):
        Context.__init__(
            self,
            components=components,
            lilypond_type=lilypond_type,
            is_simultaneous=is_simultaneous,
            name=name,
            )



import abjad
import pytest
from abjad import utilities


pytest_plugins = ['helpers_namespace']


@pytest.fixture(autouse=True)
def add_libraries(doctest_namespace):
    doctest_namespace['abjad'] = abjad
    doctest_namespace['f'] = abjad.f
    doctest_namespace['Infinity'] = abjad.mathtools.Infinity()
    doctest_namespace['NegativeInfinity'] = abjad.mathtools.NegativeInfinity()


@pytest.helpers.register
def list_all_abjad_classes(modules=None, ignored_classes=None):
    return utilities.list_all_classes(
        modules='abjad',
        ignored_classes=ignored_classes,
        )


@pytest.helpers.register
def list_all_abjad_functions(modules=None):
    return utilities.list_all_functions(modules='abjad')



def is_positive_integer_equivalent_number(argument):
    """
    Is true when ``argument`` is a positive integer-equivalent number.

    ..  container:: example

        >>> abjad.mathtools.is_positive_integer_equivalent_number(
        ...     abjad.Duration(4, 2)
        ...     )
        True

    Returns true or false.
    """
    from abjad import mathtools
    try:
        return (
            0 < argument and
            mathtools.is_integer_equivalent_number(argument)
            )
    except TypeError:  # Python 3 comparisons with non-numbers
        return False



from .Set import Set


class IntervalClassSet(Set):
    """
    Interval-class set.
    """

    ### CLASS VARIABLES ###

    __slots__ = ()

    ### INITIALIZER ###

    def __init__(self, items=None, item_class=None):
        import abjad
        prototype = (
            abjad.PitchClassSegment,
            abjad.PitchClassSet,
            abjad.PitchSegment,
            abjad.PitchSet,
            )
        if isinstance(items, prototype):
            items = list(items)
            enumerator = abjad.utilities.Enumerator(items)
            pairs = enumerator.yield_pairs()
            items = [second - first for first, second in pairs]
        Set.__init__(
            self,
            items=items,
            item_class=item_class,
            )

    ### PRIVATE PROPERTIES ###

    @property
    def _named_item_class(self):
        import abjad
        return abjad.NamedIntervalClass

    @property
    def _numbered_item_class(self):
        import abjad
        return abjad.NumberedIntervalClass

    @property
    def _parent_item_class(self):
        import abjad
        return abjad.IntervalClass

    ### PUBLIC METHODS ###

    @classmethod
    def from_selection(class_, selection, item_class=None):
        r"""
        Initialize interval set from component selection:

        ..  container:: example

            ::

                >>> staff_1 = abjad.Staff("c'4 <d' fs' a'>4 b2")
                >>> staff_2 = abjad.Staff("c4. r8 g2")
                >>> staff_group = abjad.StaffGroup([staff_1, staff_2])
                >>> abjad.show(staff_group) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(staff_group)
                \new StaffGroup
                <<
                    \new Staff
                    {
                        c'4
                        <d' fs' a'>4
                        b2
                    }
                    \new Staff
                    {
                        c4.
                        r8
                        g2
                    }
                >>

            ::

                >>> interval_classes = abjad.IntervalClassSet.from_selection(
                ...     staff_group)
                >>> for interval_class in sorted(interval_classes):
                ...     interval_class
                ...
                NamedIntervalClass('-M6')
                NamedIntervalClass('-P5')
                NamedIntervalClass('-A4')
                NamedIntervalClass('-M3')
                NamedIntervalClass('-m3')
                NamedIntervalClass('-M2')
                NamedIntervalClass('+m2')
                NamedIntervalClass('+M2')
                NamedIntervalClass('+m3')
                NamedIntervalClass('+M3')
                NamedIntervalClass('+P4')
                NamedIntervalClass('+A4')
                NamedIntervalClass('+P5')
                NamedIntervalClass('+M6')
                NamedIntervalClass('+m7')
                NamedIntervalClass('+M7')
                NamedIntervalClass('+P8')

        Returns interval set.
        """
        import abjad
        interval_set = abjad.IntervalSet.from_selection(selection)
        return class_(
            items=interval_set,
            item_class=item_class,
            )



import ide
import pathlib
import pytest
import sys


abjad_ide = ide.AbjadIDE()
scores = pathlib.Path(*pathlib.Path(__file__).parts[:-4])
path = ide.Path(__file__, scores=scores)
directories = path.materials.list_paths()


@pytest.mark.parametrize('directory', directories)
def test_materials_01(directory):
    exit_code = abjad_ide.check_definition_py(directory)
    if exit_code != 0:
        sys.exit(exit_code)


@pytest.mark.parametrize('directory', directories)
def test_materials_02(directory):
    exit_code = abjad_ide.make_illustration_pdf(directory, open_after=False)
    if exit_code != 0:
        sys.exit(exit_code)



import copy
import functools
import typing
from abjad.system.AbjadObject import AbjadObject
from abjad.lilypondnames.LilyPondTweakManager import LilyPondTweakManager
from abjad.pitch.NamedPitch import NamedPitch
from abjad.system.FormatSpecification import FormatSpecification
from abjad.system.StorageFormatManager import StorageFormatManager
from abjad.top.tweak import tweak


@functools.total_ordering
class NoteHead(AbjadObject):
    """
    Note-head.

    ..  container:: example

        >>> note = abjad.Note("cs''")
        >>> abjad.show(note) # doctest: +SKIP

        >>> note.note_head
        NoteHead("cs''")

    """

    ### CLASS VARIABLES ###

    __documentation_section__ = 'Note-heads'

    __slots__ = (
        '_alternative',
        '_client',
        '_is_cautionary',
        '_is_forced',
        '_is_parenthesized',
        '_tweaks',
        '_written_pitch',
        )

    ### INITIALIZER ###

    def __init__(
        self,
        written_pitch=None,
        client=None,
        is_cautionary=None,
        is_forced=None,
        is_parenthesized=None,
        tweaks=None,
        ):
        import abjad
        self._alternative = None
        if client is not None:
            assert isinstance(client, abjad.Leaf)
        self._client = client
        self._tweaks = None
        if isinstance(written_pitch, type(self)):
            note_head = written_pitch
            written_pitch = note_head.written_pitch
            is_cautionary = note_head.is_cautionary
            is_forced = note_head.is_forced
            tweaks = note_head.tweaks._get_attribute_pairs()
        elif written_pitch is None:
            written_pitch = 0
        self.written_pitch = written_pitch
        self.is_cautionary = is_cautionary
        self.is_forced = is_forced
        self.is_parenthesized = is_parenthesized
        LilyPondTweakManager.set_tweaks(self, tweaks)

    ### SPECIAL METHODS ###

    def __copy__(self, *arguments) -> 'NoteHead':
        """
        Copies note-head.

        ..  container:: example

            >>> import copy
            >>> note_head = abjad.NoteHead(13)
            >>> copy.copy(note_head)
            NoteHead("cs''")

        ..  container:: example

            REGRESSION. Note-heads work with new:

            >>> note = abjad.Note("cs''")
            >>> abjad.new(note.note_head)
            NoteHead("cs''")

        """
        arguments = (
            self.written_pitch,
            None,
            self.is_cautionary,
            self.is_forced,
            self.is_parenthesized,
            self.tweaks._get_attribute_pairs(),
            )
        return type(self)(*arguments)

    def __eq__(self, argument) -> bool:
        """
        Is true when ```argument`` is a note-head with written pitch equal to
        that of this note-head.
        """
        if isinstance(argument, type(self)):
            return self.written_pitch == argument.written_pitch
        return self.written_pitch == argument

    def __format__(self, format_specification='') -> str:
        """
        Formats note-head.
        """
        if format_specification in ('', 'lilypond'):
            return self._get_lilypond_format()
        elif format_specification == 'storage':
            return StorageFormatManager(self).get_storage_format()
        return str(self)

    def __hash__(self) -> int:
        """
        Hashes note-head.

        Redefined in tandem with __eq__.
        """
        return super().__hash__()

    def __lt__(self, argument) -> bool:
        """
        Is true when ``argument`` is a note-head with written pitch greater
        than that of this note-head.
        """
        if isinstance(argument, type(self)):
            return self.written_pitch < argument.written_pitch
        try:
            argument = type(self)(argument)
        except (ValueError, TypeError):
            return False
        return self.written_pitch < argument.written_pitch

    def __repr__(self) -> str:
        """
        Gets interpreter representation of note-head.

        ..  container:: example

            >>> note_head = abjad.NoteHead(13)
            >>> note_head
            NoteHead("cs''")

        """
        return super().__repr__()

    def __str__(self) -> str:
        """
        Gets string representation of note-head.

        ..  container:: example

            >>> note_head = abjad.NoteHead(13)
            >>> str(note_head)
            "cs''"

        """
        result = ''
        if self.written_pitch:
            result = str(self.written_pitch)
            if self.is_forced:
                result += '!'
            if self.is_cautionary:
                result += '?'
        return result

    ### PRIVATE PROPERTIES ###

    @property
    def _keyword_argument_names(self):
        agent = StorageFormatManager(self)
        keyword_argument_names = list(agent.signature_keyword_names)
        if 'client' in keyword_argument_names:
            keyword_argument_names.remove('client')
        if 'tweaks' in keyword_argument_names:
            keyword_argument_names.remove('tweaks')
        keyword_argument_names = tuple(keyword_argument_names)
        return keyword_argument_names

    ### PRIVATE METHODS ###

    def _get_format_specification(self):
        arguments = [repr(str(self))]
        arguments.extend(self.tweaks._get_attribute_pairs())
        arguments = ', '.join([str(x) for x in arguments])
        repr_text = f'{type(self).__name__}({arguments})'
        agent = StorageFormatManager(self)
        names = list(agent.signature_keyword_names)
        if 'client' in names:
            names.remove('client')
        if 'tweaks' in names:
            names.remove('tweaks')
        return FormatSpecification(
            self,
            repr_text=repr_text,
            storage_format_kwargs_names=names,
            )

    def _get_format_pieces(self):
        import abjad
        assert self.written_pitch
        result = []
        if self.is_parenthesized:
            result.append(r'\parenthesize')
        strings = self.tweaks._list_format_contributions(directed=False)
        result.extend(strings)
        written_pitch = self.written_pitch
        if isinstance(written_pitch, abjad.NamedPitch):
            written_pitch = written_pitch.simplify()
        kernel = format(written_pitch)
        if self.is_forced:
            kernel += '!'
        if self.is_cautionary:
            kernel += '?'
        result.append(kernel)
        return result

    def _get_lilypond_format(self, formatted_duration=None):
        import abjad
        pieces = self._get_format_pieces()
        if formatted_duration is not None:
            pieces[-1] = pieces[-1] + formatted_duration
        if self.alternative:
            pieces = abjad.LilyPondFormatManager.tag(
                pieces,
                tag=self.alternative[2],
                )
            pieces_ = self.alternative[0]._get_format_pieces()
            if formatted_duration is not None:
                pieces_[-1] = pieces_[-1] + formatted_duration
            pieces_ = abjad.LilyPondFormatManager.tag(
                pieces_,
                deactivate=True,
                tag=self.alternative[1],
                )
            pieces.extend(pieces_)
        result = '\n'.join(pieces)
        return result

    ### PUBLIC PROPERTIES ###

    @property
    def alternative(self) -> typing.Tuple['NoteHead', str, str]:
        """
        Gets and sets note-head alternative.

        ..  container:: example

            >>> note = abjad.Note("c''4")
            >>> alternative = abjad.new(note.note_head)
            >>> alternative.is_forced = True
            >>> note.note_head.alternative = (alternative, '-PARTS', '+PARTS')
            >>> abjad.show(note) # doctest: +SKIP

            >>> abjad.f(note, strict=50)
            c''4                                              %! +PARTS
            %@% c''!4                                         %! -PARTS

            Survives pitch reassignment:

            >>> note.written_pitch = 'D5'
            >>> abjad.show(note) # doctest: +SKIP

            >>> abjad.f(note, strict=50)
            d''4                                              %! +PARTS
            %@% d''!4                                         %! -PARTS

            Clear with none:

            >>> note.note_head.alternative = None
            >>> abjad.show(note) # doctest: +SKIP

            >>> abjad.f(note, strict=50)
            d''4

        ..  container:: example

            >>> chord = abjad.Chord("<c' d' bf''>4")
            >>> alternative = abjad.new(chord.note_heads[0])
            >>> alternative.is_forced = True
            >>> chord.note_heads[0].alternative = (alternative, '-PARTS', '+PARTS')
            >>> abjad.show(chord) # doctest: +SKIP

            >>> abjad.f(chord, strict=50)
            <
                c'                                            %! +PARTS
            %@% c'!                                           %! -PARTS
                d'
                bf''
            >4

            Suvives pitch reassignment:

            >>> chord.note_heads[0].written_pitch = 'B3'
            >>> abjad.show(chord) # doctest: +SKIP

            >>> abjad.f(chord, strict=50)
            <
                b                                             %! +PARTS
            %@% b!                                            %! -PARTS
                d'
                bf''
            >4

            Clear with none:

            >>> chord.note_heads[0].alternative = None
            >>> abjad.f(chord, strict=50)
            <b d' bf''>4

        """
        return self._alternative

    @alternative.setter
    def alternative(self, argument):
        if argument is not None:
            assert isinstance(argument, tuple), repr(argument)
            assert len(argument) == 3, repr(argument)
            assert isinstance(argument[0], NoteHead), repr(argument)
            assert argument[0].alternative is None, repr(argument)
            assert isinstance(argument[1], str), repr(argument)
            assert isinstance(argument[2], str), repr(argument)
        self._alternative = argument

    @property
    def client(self):
        """
        Gets client of note-head.

        ..  container:: example

            >>> note_head = abjad.NoteHead(13)
            >>> note_head.client is None
            True

        Returns note, chord or none.
        """
        return self._client

    @property
    def is_cautionary(self) -> bool:
        """
        Gets and sets cautionary accidental flag.

        ..  container:: example

            >>> note = abjad.Note("c''")
            >>> note.note_head.is_cautionary = True
            >>> abjad.show(note) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(note)
                c''?4

            >>> note = abjad.Note("cs''")
            >>> note.note_head.is_cautionary = True
            >>> abjad.show(note) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(note)
                cs''?4

        """
        return self._is_cautionary

    @is_cautionary.setter
    def is_cautionary(self, argument):
        if argument is not None:
            argument = bool(argument)
        self._is_cautionary = argument

    @property
    def is_forced(self) -> bool:
        """
        Gets and sets forced accidental flag.

        ..  container:: example

            >>> note = abjad.Note("c''")
            >>> note.note_head.is_forced = True
            >>> abjad.show(note) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(note)
                c''!4

            >>> note = abjad.Note("cs''")
            >>> note.note_head.is_forced = True
            >>> abjad.show(note) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(note)
                cs''!4

        """
        return self._is_forced

    @is_forced.setter
    def is_forced(self, argument):
        if argument is not None:
            argument = bool(argument)
        self._is_forced = argument

    @property
    def is_parenthesized(self) -> bool:
        r"""
        Gets and sets forced accidental flag.

        ..  container:: example

            >>> note = abjad.Note("c''")
            >>> note.note_head.is_parenthesized = True
            >>> abjad.show(note) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(note)
                \parenthesize
                c''4

            >>> note = abjad.Note("cs''")
            >>> note.note_head.is_parenthesized = True
            >>> abjad.show(note) # doctest: +SKIP

            ..  docs::

                >>> abjad.f(note)
                \parenthesize
                cs''4

        """
        return self._is_parenthesized

    @is_parenthesized.setter
    def is_parenthesized(self, argument):
        if argument is not None:
            argument = bool(argument)
        self._is_parenthesized = argument

    @property
    def named_pitch(self) -> NamedPitch:
        """
        Gets named pitch.

        ..  container:: example

            >>> note_head = abjad.NoteHead("cs''")
            >>> note_head.named_pitch
            NamedPitch("cs''")

        """
        return self.written_pitch

    @property
    def tweaks(self) -> LilyPondTweakManager:
        r"""
        Gets tweaks.

        ..  container:: example

            >>> note_head = abjad.NoteHead("cs''")
            >>> note_head.tweaks
            LilyPondTweakManager()

            >>> abjad.tweak(note_head).color = 'red'
            >>> abjad.f(note_head)
            \tweak color #red
            cs''

        """
        return tweak(self)

    @property
    def written_pitch(self) -> NamedPitch:
        """
        Gets and sets written pitch of note-head.

        ..  container:: example

            >>> note_head = abjad.NoteHead("cs''")
            >>> note_head.written_pitch
            NamedPitch("cs''")

            >>> note_head = abjad.NoteHead("cs''")
            >>> note_head.written_pitch = "d''"
            >>> note_head.written_pitch
            NamedPitch("d''")

        """
        return self._written_pitch

    @written_pitch.setter
    def written_pitch(self, argument):
        written_pitch = NamedPitch(argument)
        self._written_pitch = written_pitch
        if self.alternative is not None:
            self.alternative[0].written_pitch = written_pitch



from abjad import markups
from abjad.system.AbjadValueObject import AbjadValueObject


class Rotation(AbjadValueObject):
    """
    Rotation operator.

    ..  container:: example:

        >>> abjad.Rotation()
        Rotation(n=0)

    ..  container:: example

        >>> abjad.Rotation(n=1)
        Rotation(n=1)

    Object model of the twelve-tone rotation operator.
    """

    ### CLASS VARIABLES ###

    __slots__ = (
        '_n',
        '_period',
        '_stravinsky',
        )

    ### INITIALIZER ###

    def __init__(self, *, n=0, period=None, stravinsky=None):
        self._n = int(n)
        if period is not None:
            period = abs(int(period))
            assert 0 < period
        self._period = period
        assert isinstance(stravinsky, (bool, type(None))), repr(stravinsky)
        self._stravinsky = stravinsky

    ### SPECIAL METHODS ###

    def __add__(self, operator):
        r"""
        Composes rotation and `operator`.

        ..  container:: example

            Example segment:

            >>> items = [0, 2, 4, 5]
            >>> segment = abjad.PitchClassSegment(items=items)
            >>> abjad.show(segment) # doctest: +SKIP

            Example operators:

            >>> rotation = abjad.Rotation(n=-1)
            >>> transposition = abjad.Transposition(n=3)

        ..  container:: example

            Transposition followed by rotation:

            >>> operator = rotation + transposition
            >>> str(operator)
            'r-1T3'

            >>> segment_ = operator(segment)
            >>> abjad.show(segment_) # doctest: +SKIP

            ..  docs::

                >>> lilypond_file = segment_.__illustrate__()
                >>> abjad.f(lilypond_file[abjad.Voice])
                \new Voice
                {
                    f'8
                    g'8
                    af'8
                    ef'8
                    \bar "|." %! SCORE1
                    \override Score.BarLine.transparent = ##f
                }

        ..  container:: example

            Same as above because rotation and transposition commute:

            >>> operator = transposition + rotation
            >>> str(operator)
            'T3r-1'

            >>> segment_ = operator(segment)
            >>> abjad.show(segment_) # doctest: +SKIP

            ..  docs::

                >>> lilypond_file = segment_.__illustrate__()
                >>> abjad.f(lilypond_file[abjad.Voice])
                \new Voice
                {
                    f'8
                    g'8
                    af'8
                    ef'8
                    \bar "|." %! SCORE1
                    \override Score.BarLine.transparent = ##f
                }

        ..  container:: example

            Returns compound operator:

            >>> abjad.f(operator)
            abjad.CompoundOperator(
                operators=[
                    abjad.Rotation(
                        n=-1,
                        ),
                    abjad.Transposition(
                        n=3,
                        ),
                    ],
                )

        """
        import abjad
        return abjad.CompoundOperator._compose_operators(self, operator)

    def __call__(self, argument):
        """
        Calls rotation on `argument`.

        ..  container:: example

            Rotates pitch classes:

            >>> rotation = abjad.Rotation(n=1)
            >>> pitch_classes = abjad.PitchClassSegment([0, 1, 4, 7])
            >>> rotation(pitch_classes)
            PitchClassSegment([7, 0, 1, 4])

        ..  container:: example

            Rotates pitch classes with Stravinsky-style back-transposition to
            zero:

            >>> rotation = abjad.Rotation(n=1, stravinsky=True)
            >>> pitch_classes = abjad.PitchClassSegment([0, 1, 4, 7])
            >>> rotation(pitch_classes)
            PitchClassSegment([0, 5, 6, 9])

        ..  container:: example

            Does not rotate single pitches or pitch-classes:

            >>> rotation = abjad.Rotation(n=1)
            >>> pitch_class = abjad.NumberedPitchClass(6)
            >>> rotation(pitch_class)
            NumberedPitchClass(6)

        ..  container:: example

            Periodic rotation:

            ..  todo:: Deprecated.

            >>> rotation = abjad.Rotation(n=1, period=3)
            >>> pitches = abjad.PitchSegment("c' d' e' f' g' a' b' c''")
            >>> rotation(pitches)
            PitchSegment("e' c' d' a' f' g' c'' b'")

        ..  container:: example

            Stravinsky-style periodic rotation:

            ..  todo:: Deprecated.

            >>> rotation = abjad.Rotation(
            ...     n=1,
            ...     period=3,
            ...     stravinsky=True,
            ...     )
            >>> pitches = abjad.PitchSegment("c' d' e' f' g' a' b' c''")
            >>> rotation(pitches)
            PitchSegment("c' af bf f' df' ef' b' as'")

        Returns new object with type equal to that of `argument`.
        """
        import abjad
        if isinstance(argument, (abjad.Pitch, abjad.PitchClass)):
            return argument
        if not isinstance(argument, (
            abjad.PitchSegment,
            abjad.PitchClassSegment,
            )):
            argument = abjad.PitchSegment(argument)
        if not self.period:
            return argument.rotate(self.n, stravinsky=self.stravinsky)
        result = abjad.new(argument, items=())
        for shard in abjad.sequence(argument).partition_by_counts(
            [self.period],
            cyclic=True,
            overhang=True,
            ):
            shard = type(argument)(shard)
            shard = shard.rotate(self.n, stravinsky=self.stravinsky)
            result = result + shard
        return result

    def __radd__(self, operator):
        """
        Right-addition not defined on rotation.

        ..  container:: example

            >>> abjad.Rotation().__radd__(abjad.Rotation())
            Traceback (most recent call last):
                ...
            NotImplementedError: right-addition not defined on Rotation.

        Raises not implemented error.
        """
        message = 'right-addition not defined on {}.'
        message = message.format(type(self).__name__)
        raise NotImplementedError(message)

    def __str__(self):
        """
        Gets string representation of operator.

        ..  container:: example

            >>> str(abjad.Rotation())
            'r0'

        ..  container:: example

            >>> str(abjad.Rotation(n=1))
            'r1'

        ..  container:: example

            >>> str(abjad.Rotation(stravinsky=True))
            'rs0'

        ..  container:: example

            >>> str(abjad.Rotation(n=1, stravinsky=True))
            'rs1'

        """
        string = 'r{}'
        if self.stravinsky:
            string = 'rs{}'
        string = string.format(self.n)
        return string

    ### PRIVATE METHODS ###

    def _get_markup(self, direction=None):
        operator = markups.Markup('r', direction=direction)
        subscript = markups.Markup(self.n).sub()
        hspace = markups.Markup.hspace(-0.25)
        markup = markups.Markup.concat([operator, hspace, subscript])
        return markup

    def _is_identity_operator(self):
        if self.n == 0:
            return True
        return False

    ### PUBLIC PROPERTIES ###

    @property
    def n(self):
        """
        Gets index of rotation.

        ..  container:: example

            >>> rotation = abjad.Rotation()
            >>> rotation.n
            0

        ..  container:: example

            >>> rotation = abjad.Rotation(n=2)
            >>> rotation.n
            2

        Returns integer.
        """
        return self._n

    @property
    def period(self):
        """
        Gets period of rotation.

        ..  todo:: Deprecated.

        ..  container:: example

            >>> rotation = abjad.Rotation(n=2, period=3)
            >>> rotation.period
            3

        Returns integer or none.
        """
        return self._period

    @property
    def stravinsky(self):
        """
        Is true when rotation uses Stravinsky-style back-transposition to zero.

        ..  container:: example

            >>> rotation = abjad.Rotation(n=2, stravinsky=False)
            >>> rotation.stravinsky
            False

        Returns true or false.
        """
        return self._stravinsky



from abjad import enums
from abjad import exceptions
from abjad.system.AbjadObject import AbjadObject


class UpdateManager(AbjadObject):
    """
    Update manager.

    Updates start offset, stop offsets and indicators everywhere in score.
    """

    ### CLASS VARIABLES ###

    __documentation_section__ = 'Managers'

    __slots__ = ()

    ### PRIVATE METHODS ###

    @staticmethod
    def _get_score_tree_state_flags(parentage):
        offsets_are_current = True
        indicators_are_current = True
        offsets_in_seconds_are_current = True
        for component in parentage:
            if offsets_are_current:
                if not component._offsets_are_current:
                    offsets_are_current = False
            if indicators_are_current:
                if not component._indicators_are_current:
                    indicators_are_current = False
            if offsets_in_seconds_are_current:
                if not component._offsets_in_seconds_are_current:
                    offsets_in_seconds_are_current = False
        return (
            offsets_are_current,
            indicators_are_current,
            offsets_in_seconds_are_current,
            )

    @staticmethod
    def _iterate_entire_score(score_root):
        import abjad
        components = abjad.iterate(score_root)._depth_first(
            capped=True,
            unique=True,
            forbid=None,
            direction=enums.Left,
            )
        return components

    def _make_metronome_mark_map(self, score_root):
        from abjad.utilities.Multiplier import Multiplier
        from abjad.utilities.Offset import Offset
        from abjad.utilities.Sequence import Sequence
        from abjad.indicators.MetronomeMark import MetronomeMark
        from abjad.timespans.AnnotatedTimespan import AnnotatedTimespan
        from abjad.timespans.TimespanList import TimespanList
        pairs = []
        all_stop_offsets = set()
        for component in self._iterate_entire_score(score_root):
            indicators = component._get_indicators(MetronomeMark)
            if len(indicators) == 1:
                metronome_mark = indicators[0]
                if not metronome_mark.is_imprecise:
                    pair = (component._start_offset, metronome_mark)
                    pairs.append(pair)
            if component._stop_offset is not None:
                all_stop_offsets.add(component._stop_offset)
        pairs.sort(key=lambda _:_[0])
        if not pairs:
            return
        if pairs[0][0] != 0:
            return
        score_stop_offset = max(all_stop_offsets)
        timespans = TimespanList()
        clocktime_rate = MetronomeMark((1, 4), 60)
        clocktime_start_offset = Offset(0)
        for left, right in Sequence(pairs).nwise(wrapped=True):
            metronome_mark = left[-1]
            start_offset = left[0]
            stop_offset = right[0]
            # last timespan
            if stop_offset == 0:
                stop_offset = score_stop_offset
            duration = stop_offset - start_offset
            multiplier = Multiplier(60, metronome_mark.units_per_minute)
            clocktime_duration = duration / metronome_mark.reference_duration
            clocktime_duration *= multiplier
            timespan = AnnotatedTimespan(
                start_offset=start_offset,
                stop_offset=stop_offset,
                annotation=(clocktime_start_offset, clocktime_duration),
                )
            timespans.append(timespan)
            clocktime_start_offset += clocktime_duration
        return timespans

    def _update_all_indicators(self, score_root):
        """
        Updating indicators does not update offsets.
        On the other hand, getting an effective indicator does update
        offsets when at least one indicator of the appropriate type
        attaches to score.
        """
        import abjad
        components = self._iterate_entire_score(score_root)
        for component in components:
            for wrapper in abjad.inspect(component).wrappers():
                if wrapper.context is not None:
                    wrapper._update_effective_context()
            component._indicators_are_current = True

    @staticmethod
    def _update_all_leaf_indices_and_measure_numbers(score_root):
        """
        Call only when updating offsets.
        No separate state flags exist for leaf indices or measure numbers.
        """
        import abjad
        from abjad.top import iterate
        if isinstance(score_root, abjad.Context):
            contexts = iterate(score_root).components(abjad.Context)
            for context in contexts:
                for leaf_index, leaf in enumerate(iterate(context).leaves()):
                    leaf._leaf_index = leaf_index
                for measure_index, measure in enumerate(
                    iterate(context).components(abjad.Measure)):
                    measure_number = measure_index + 1
                    measure._measure_number = measure_number
        else:
            for leaf_index, leaf in enumerate(iterate(score_root).leaves()):
                leaf._leaf_index = leaf_index
            for measure_index, measure in enumerate(
                iterate(score_root).components(abjad.Measure)):
                measure_number = measure_index + 1
                measure._measure_number = measure_number

    def _update_all_offsets(self, score_root):
        """
        Updating offsets does not update indicators.
        Updating offsets does not update offsets in seconds.
        """
        for component in self._iterate_entire_score(score_root):
            self._update_component_offsets(component)
            component._offsets_are_current = True

    def _update_all_offsets_in_seconds(self, score_root):
        self._update_all_offsets(score_root)
        timespans = self._make_metronome_mark_map(score_root)
        for component in self._iterate_entire_score(score_root):
            self._update_clocktime_offsets(component, timespans)
            component._offsets_in_seconds_are_current = True

    @staticmethod
    def _update_clocktime_offsets(component, timespans):
        from abjad.utilities.Offset import Offset
        if not timespans:
            return
        for timespan in timespans:
            if (timespan.start_offset <= component._start_offset <
                timespan.stop_offset):
                pair = timespan.annotation
                clocktime_start_offset, clocktime_duration = pair
                local_offset = component._start_offset - timespan.start_offset
                multiplier = local_offset / timespan.duration
                duration = multiplier * clocktime_duration
                offset = clocktime_start_offset + duration
                component._start_offset_in_seconds = Offset(offset)
            if (timespan.start_offset <= component._stop_offset <
                timespan.stop_offset):
                pair = timespan.annotation
                clocktime_start_offset, clocktime_duration = pair
                local_offset = component._stop_offset - timespan.start_offset
                multiplier = local_offset / timespan.duration
                duration = multiplier * clocktime_duration
                offset = clocktime_start_offset + duration
                component._stop_offset_in_seconds = Offset(offset)
                return
        if component._stop_offset == timespans[-1].stop_offset:
            pair = timespans[-1].annotation
            clocktime_start_offset, clocktime_duration = pair
            offset = clocktime_start_offset + clocktime_duration
            component._stop_offset_in_seconds = Offset(offset)
            return
        raise Exception(f'can not find {stop_offset} in {timespans}.')

    @classmethod
    def _update_component_offsets(class_, component):
        import abjad
        if isinstance(component._parent, abjad.GraceContainer):
            pair = class_._get_grace_note_offsets(component)
            start_offset, stop_offset = pair
        elif isinstance(component._parent, abjad.AfterGraceContainer):
            pair = class_._get_after_grace_note_offsets(component)
            start_offset, stop_offset = pair
        else:
            previous = component._get_nth_component_in_time_order_from(-1)
            if previous is not None:
                start_offset = previous._stop_offset
            else:
                start_offset = abjad.Offset(0)
            stop_offset = start_offset + component._get_duration()
        component._start_offset = start_offset
        component._stop_offset = stop_offset
        component._timespan._start_offset = start_offset
        component._timespan._stop_offset = stop_offset

    def _update_now(
        self,
        component,
        offsets=False,
        offsets_in_seconds=False,
        indicators=False,
        ):
        import abjad
        assert offsets or offsets_in_seconds or indicators
        if component._is_forbidden_to_update:
            return
        parentage = abjad.inspect(component).parentage(
            include_self=True,
            grace_notes=True,
            )
        for parent in parentage:
            if parent._is_forbidden_to_update:
                return
        (
            offsets_are_current,
            indicators_are_current,
            offsets_in_seconds_are_current,
            ) = self._get_score_tree_state_flags(parentage)
        score_root = parentage.root
        if offsets and not offsets_are_current:
            self._update_all_offsets(score_root)
            self._update_all_leaf_indices_and_measure_numbers(score_root)
        if offsets_in_seconds and not offsets_in_seconds_are_current:
            self._update_all_offsets_in_seconds(score_root)
        if indicators and not indicators_are_current:
            self._update_all_indicators(score_root)

    ### EXPERIMENTAL ###

    @staticmethod
    def _get_after_grace_note_offsets(grace_note):
        import abjad
        after_grace_container = grace_note._parent
        assert isinstance(after_grace_container, abjad.AfterGraceContainer)
        carrier_leaf = after_grace_container._carrier
        carrier_leaf_stop_offset = carrier_leaf._stop_offset
        grace_displacement = -grace_note.written_duration
        sibling = grace_note._get_sibling(1)
        while sibling is not None:
            grace_displacement -= sibling.written_duration
            sibling = sibling._get_sibling(1)
        start_offset = abjad.Offset(
            carrier_leaf_stop_offset,
            grace_displacement=grace_displacement,
            )
        grace_displacement += grace_note.written_duration
        stop_offset = abjad.Offset(
            carrier_leaf_stop_offset,
            grace_displacement=grace_displacement,
            )
        return start_offset, stop_offset

    @staticmethod
    def _get_grace_note_offsets(grace_note):
        import abjad
        grace_container = grace_note._parent
        carrier_leaf = grace_container._carrier
        carrier_leaf_start_offset = carrier_leaf._start_offset
        grace_displacement = -grace_note.written_duration
        sibling = grace_note._get_sibling(1)
        while sibling is not None:
            grace_displacement -= sibling.written_duration
            sibling = sibling._get_sibling(1)
        start_offset = abjad.Offset(
            carrier_leaf_start_offset,
            grace_displacement=grace_displacement,
            )
        grace_displacement += grace_note.written_duration
        stop_offset = abjad.Offset(
            carrier_leaf_start_offset,
            grace_displacement=grace_displacement,
            )
        return start_offset, stop_offset

    def _get_measure_start_offsets(self, component):
        import abjad
        wrappers = []
        prototype = abjad.TimeSignature
        score_root = abjad.inspect(component).parentage(
            include_self=True).root
        for component in self._iterate_entire_score(score_root):
            wrappers_ = abjad.inspect(component).wrappers(prototype)
            wrappers.extend(wrappers_)
        pairs = []
        for wrapper in wrappers:
            inspector = abjad.inspect(wrapper.component)
            start_offset = inspector.timespan().start_offset
            time_signature = wrapper.indicator
            pair = start_offset, time_signature
            pairs.append(pair)
        offset_zero = abjad.Offset(0)
        default_time_signature = abjad.TimeSignature((4, 4))
        default_pair = (offset_zero, default_time_signature)
        if pairs and not pairs[0] == offset_zero:
            pairs.insert(0, default_pair)
        elif not pairs:
            pairs = [default_pair]
        pairs.sort(key=lambda x: x[0])
        parentage = abjad.inspect(component).parentage()
        score_root = parentage.root
        inspector = abjad.inspect(score_root)
        score_stop_offset = inspector.timespan().stop_offset
        dummy_last_pair = (score_stop_offset, None)
        pairs.append(dummy_last_pair)
        measure_start_offsets = []
        for current_pair, next_pair in abjad.sequence(pairs).nwise():
            current_start_offset, current_time_signature = current_pair
            next_start_offset, next_time_signature = next_pair
            measure_start_offset = current_start_offset
            while measure_start_offset < next_start_offset:
                measure_start_offsets.append(measure_start_offset)
                measure_start_offset += current_time_signature.duration
        return measure_start_offsets

    # TODO: reimplement with some type of bisection
    def _to_measure_number(
        self,
        component,
        measure_number_start_offsets,
        ):
        import abjad
        inspector = abjad.inspect(component)
        component_start_offset = inspector.timespan().start_offset
        measure_number_start_offsets = measure_number_start_offsets[:]
        measure_number_start_offsets.append(abjad.mathtools.Infinity())
        pairs = abjad.sequence(measure_number_start_offsets)
        pairs = pairs.nwise()
        for measure_index, pair in enumerate(pairs):
            if pair[0] <= component_start_offset < pair[-1]:
                measure_number = measure_index + 1
                return measure_number
        message = 'can not find measure number: {!r}, {!r}.'
        message = message.format(component, measure_number_start_offsets)
        raise ValueError(message)

    def _update_measure_numbers(self, component):
        import abjad
        measure_start_offsets = self._get_measure_start_offsets(component)
        assert measure_start_offsets, repr(measure_start_offsets)
        score_root = abjad.inspect(component).parentage(
            include_self=True).root
        for component in self._iterate_entire_score(score_root):
            measure_number = self._to_measure_number(
                component,
                measure_start_offsets,
                )
            component._measure_number = measure_number
